# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import numpy as np

from copy import deepcopy

from xypattern import Pattern
from xypattern.pattern import SmoothBrucknerBackground

from .util import Signal
from .util.ImgCorrection import CbnCorrection, ObliqueAngleDetectorAbsorptionCorrection

from .util.calc import convert_units
from . import ImgModel, CalibrationModel, MaskModel, PatternModel, BatchModel
from .MapModel2 import MapModel2
from .CalibrationModel import DetectorModes


class Configuration(object):
    """
    The configuration class contains a working combination of an ImgModel, PatternModel, MaskModel and CalibrationModel.
    It does handles the core data manipulation of Dioptas.
    The management of multiple Configurations is done by the DioptasModel.
    """

    def __init__(self, working_directories=None):
        super(Configuration, self).__init__()

        self.img_model = ImgModel()
        self.mask_model = MaskModel()
        self.calibration_model = CalibrationModel(self.img_model)
        self.pattern_model = PatternModel()

        self.batch_model = BatchModel(self)
        self.map_model = MapModel2(self)

        if working_directories is None:
            self.working_directories = {
                "calibration": "",
                "mask": "",
                "image": os.path.expanduser("~"),
                "pattern": "",
                "overlay": "",
                "phase": "",
                "batch": os.path.expanduser("~"),
            }
        else:
            self.working_directories = working_directories

        self.use_mask = False

        self.transparent_mask = False

        self._integration_rad_points = None
        self._integration_unit = "2th_deg"
        self._oned_azimuth_range = None
        self.trim_trailing_zeros = True

        self._cake_azimuth_points = 360
        self._cake_azimuth_range = None

        self._auto_integrate_pattern = True
        self._auto_integrate_cake = False

        self.auto_save_integrated_pattern = False
        self.integrated_patterns_file_formats = [".xy"]

        self.cake_changed = Signal()
        self._connect_signals()

    def _connect_signals(self):
        """
        Connects the img_changed signal to responding functions.
        """
        self.img_model.img_changed.connect(self.update_mask_dimension)
        self.img_model.img_changed.connect(self.integrate_image_1d)

    def integrate_image_1d(self):
        """
        Integrates the image in the ImageModel to a Pattern. Will also automatically save the integrated pattern, if
        auto_save_integrated is True.
        """
        if self.calibration_model.is_calibrated:
            if self.use_mask:
                mask = self.mask_model.get_mask()
            elif self.mask_model.roi is not None:
                mask = self.mask_model.roi_mask
            else:
                mask = None

            x, y = self.calibration_model.integrate_1d(
                azi_range=self.oned_azimuth_range,
                mask=mask,
                unit=self.integration_unit,
                num_points=self.integration_rad_points,
                trim_zeros=self.trim_trailing_zeros,
            )

            self.pattern_model.set_pattern(
                x, y, self.img_model.filename, unit=self.integration_unit
            )  #

            if self.auto_save_integrated_pattern:
                self._auto_save_patterns()

            return x, y

    def integrate_image_2d(self):
        """
        Integrates the image in the ImageModel to a Cake.
        """
        if self.use_mask:
            mask = self.mask_model.get_mask()
        elif self.mask_model.roi is not None:
            mask = self.mask_model.roi_mask
        else:
            mask = None

        self.calibration_model.integrate_2d(
            mask=mask,
            rad_points=self._integration_rad_points,
            azimuth_points=self._cake_azimuth_points,
            azimuth_range=self._cake_azimuth_range,
        )

        self.cake_changed.emit()

    def save_pattern(self, filename=None, subtract_background=False):
        """
        Saves the current integrated pattern. The format depends on the file ending. Possible file formats:
            [*.xy, *.chi, *.dat, *.fxye]
        :param filename: where to save the file
        :param subtract_background: flat whether the pattern should be saved with or without subtracted background
        """
        if filename is None:
            filename = self.img_model.filename

        if filename.endswith(".xy"):
            self.pattern_model.save_pattern(
                filename,
                header=self._create_xy_header(),
                subtract_background=subtract_background,
            )
        elif filename.endswith(".fxye"):
            self.pattern_model.save_pattern(
                filename,
                header=self._create_fxye_header(filename),
                subtract_background=subtract_background,
            )
        else:
            self.pattern_model.save_pattern(
                filename, subtract_background=subtract_background
            )

    def save_background_pattern(self, filename=None):
        """
        Saves the current fit background as a pattern. The format depends on the file ending. Possible file formats:
            [*.xy, *.chi, *.dat, *.fxye]
        """
        if filename is None:
            filename = self.img_model.filename

        if filename.endswith(".xy"):
            self.pattern_model.save_auto_background_as_pattern(
                filename, header=self._create_xy_header()
            )
        elif filename.endswith(".fxye"):
            self.pattern_model.save_auto_background_as_pattern(
                filename, header=self._create_fxye_header(filename)
            )
        else:
            self.pattern_model.save_pattern(filename)

    def _create_xy_header(self) -> str:
        """
        Creates the header for the xy file format (contains information about calibration parameters).
        :return: header string
        """
        header = self.calibration_model.create_file_header()
        header = header.replace("\r\n", "\n")
        header = header + "\n#\n# " + self._integration_unit + "\t I"
        return header

    def _create_fxye_header(self, filename) -> str:
        """
        Creates the header for the fxye file format (used by GSAS and GSAS-II) containing the calibration information
        :return: header string
        """
        header = "Generated file " + filename + " using DIOPTAS\n"
        header = header + self.calibration_model.create_file_header()
        unit = self._integration_unit
        lam = self.calibration_model.wavelength
        if unit == "q_A^-1":
            con = "CONQ"
        else:
            con = "CONS"

        header = (
            header
            + "\nBANK\t1\tNUM_POINTS\tNUM_POINTS "
            + con
            + "\tMIN_X_VAL\tSTEP_X_VAL "
            + "{0:.5g}".format(lam * 1e10)
            + " 0.0 FXYE"
        )
        return header

    def _auto_save_patterns(self):
        """
        Saves the current pattern in the pattern working directory (specified in self.working_directories['pattern'].
        When background subtraction is enabled in the pattern model the pattern will be saved with background
        subtraction and without in another sub-folder. ('bkg_subtracted')
        """
        for file_ending in self.integrated_patterns_file_formats:
            filename = os.path.join(
                self.working_directories["pattern"],
                os.path.basename(str(self.img_model.filename)).split(".")[:-1][0]
                + file_ending,
            )
            filename = filename.replace("\\", "/")
            self.save_pattern(filename)

        if self.pattern_model.pattern.has_background():
            for file_ending in self.integrated_patterns_file_formats:
                directory = os.path.join(
                    self.working_directories["pattern"], "bkg_subtracted"
                )
                if not os.path.exists(directory):
                    os.mkdir(directory)
                filename = os.path.join(
                    directory, self.pattern_model.pattern.name + file_ending
                )
                filename = filename.replace("\\", "/")
                self.save_pattern(filename, subtract_background=True)

    def update_mask_dimension(self):
        """
        Updates the shape of the mask in the MaskModel to the shape of the image in the ImageModel.
        """
        self.mask_model.set_dimension(self.img_model._img_data.shape)

    @property
    def integration_rad_points(self) -> int:
        return self._integration_rad_points

    @integration_rad_points.setter
    def integration_rad_points(self, new_value: int):
        self._integration_rad_points = new_value
        self.integrate_image_1d()
        if self.auto_integrate_cake:
            self.integrate_image_2d()

    @property
    def cake_azimuth_points(self) -> int:
        return self._cake_azimuth_points

    @cake_azimuth_points.setter
    def cake_azimuth_points(self, new_value):
        self._cake_azimuth_points = new_value
        if self.auto_integrate_cake:
            self.integrate_image_2d()

    @property
    def cake_azimuth_range(self):
        return self._cake_azimuth_range

    @cake_azimuth_range.setter
    def cake_azimuth_range(self, new_value):
        self._cake_azimuth_range = new_value
        if self.auto_integrate_cake:
            self.integrate_image_2d()

    @property
    def oned_azimuth_range(self):
        return self._oned_azimuth_range

    @oned_azimuth_range.setter
    def oned_azimuth_range(self, new_value):
        self._oned_azimuth_range = new_value
        if self.auto_integrate_pattern:
            self.integrate_image_1d()

    @property
    def integration_unit(self) -> str:
        return self._integration_unit

    @integration_unit.setter
    def integration_unit(self, new_unit: str):
        old_unit = self.integration_unit
        self._integration_unit = new_unit

        self.pattern_model.pattern.transform_x(
            lambda x: convert_units(
                x, self.calibration_model.wavelength, old_unit, new_unit
            )
        )

        self.integrate_image_1d()

    @property
    def correct_solid_angle(self):
        return self.calibration_model.correct_solid_angle

    @correct_solid_angle.setter
    def correct_solid_angle(self, new_val):
        self.calibration_model.correct_solid_angle = new_val
        if self.auto_integrate_pattern:
            self.integrate_image_1d()
        if self._auto_integrate_cake:
            self.integrate_image_2d()

    @property
    def is_calibrated(self) -> bool:
        return self.calibration_model.is_calibrated

    @property
    def auto_integrate_cake(self) -> bool:
        return self._auto_integrate_cake

    @auto_integrate_cake.setter
    def auto_integrate_cake(self, new_value):
        if self._auto_integrate_cake == new_value:
            return

        self._auto_integrate_cake = new_value
        if new_value:
            self.img_model.img_changed.connect(self.integrate_image_2d)
        else:
            self.img_model.img_changed.disconnect(self.integrate_image_2d)

    @property
    def auto_integrate_pattern(self) -> bool:
        return self._auto_integrate_pattern

    @auto_integrate_pattern.setter
    def auto_integrate_pattern(self, new_value):
        if self._auto_integrate_pattern == new_value:
            return

        self._auto_integrate_pattern = new_value
        if new_value:
            self.img_model.img_changed.connect(self.integrate_image_1d)
        else:
            self.img_model.img_changed.disconnect(self.integrate_image_1d)

    @property
    def cake_img(self) -> np.ndarray:
        return self.calibration_model.cake_img

    @property
    def roi(self):
        return self.mask_model.roi

    @roi.setter
    def roi(self, new_val):
        self.mask_model.roi = new_val
        self.integrate_image_1d()

    def copy(self) -> "Configuration":
        """
        Creates a copy of the current configuration directory
        :return: copied configuration
        :rtype: Configuration
        """
        new_configuration = Configuration(self.working_directories)
        new_configuration.img_model._img_data = self.img_model._img_data
        new_configuration.img_model.img_transformations = deepcopy(
            self.img_model.img_transformations
        )

        new_configuration.calibration_model.set_pyFAI(
            self.calibration_model.get_calibration_parameter()[0]
        )
        new_configuration.integrate_image_1d()

        return new_configuration

    def save_in_hdf5(self, hdf5_group):
        """
        Saves the configuration group in the given hdf5_group.
        :type hdf5_group: h5py.Group
        """

        f = hdf5_group
        # save general information
        general_information = f.create_group("general_information")
        # integration parameters:
        general_information.attrs["integration_unit"] = self.integration_unit
        if self.integration_rad_points:
            general_information.attrs["integration_num_points"] = (
                self.integration_rad_points
            )
        else:
            general_information.attrs["integration_num_points"] = 0

        # cake parameters:
        general_information.attrs["auto_integrate_cake"] = self.auto_integrate_cake
        general_information.attrs["cake_azimuth_points"] = self.cake_azimuth_points
        if self.cake_azimuth_range is None:
            general_information.attrs["cake_azimuth_range"] = "None"
        else:
            general_information.attrs["cake_azimuth_range"] = self.cake_azimuth_range

        # mask parameters
        general_information.attrs["use_mask"] = self.use_mask
        general_information.attrs["transparent_mask"] = self.transparent_mask

        # auto save parameters
        general_information.attrs["auto_save_integrated_pattern"] = (
            self.auto_save_integrated_pattern
        )
        formats = [
            n.encode("ascii", "ignore") for n in self.integrated_patterns_file_formats
        ]
        general_information.create_dataset(
            "integrated_patterns_file_formats", (len(formats), 1), "S10", formats
        )

        # save working directories
        working_directories_gp = f.create_group("working_directories")
        try:
            for key in self.working_directories:
                working_directories_gp.attrs[key] = self.working_directories[key]
        except TypeError:
            self.working_directories = {
                "calibration": "",
                "mask": "",
                "image": "",
                "pattern": "",
                "overlay": "",
                "phase": "",
                "batch": "",
            }
            for key in self.working_directories:
                working_directories_gp.attrs[key] = self.working_directories[key]

        # save image model
        image_group = f.create_group("image_model")
        image_group.attrs["auto_process"] = self.img_model.autoprocess
        image_group.attrs["factor"] = self.img_model.factor
        image_group.attrs["has_background"] = self.img_model.has_background()
        image_group.attrs["background_filename"] = self.img_model.background_filename
        image_group.attrs["background_offset"] = self.img_model.background_offset
        image_group.attrs["background_scaling"] = self.img_model.background_scaling
        if self.img_model.has_background():
            background_data = self.img_model.untransformed_background_data
            image_group.create_dataset(
                "background_data", background_data.shape, "f", background_data
            )

        image_group.attrs["series_max"] = self.img_model.series_max
        image_group.attrs["series_pos"] = self.img_model.series_pos

        # image corrections
        corrections_group = image_group.create_group("corrections")
        corrections_group.attrs["has_corrections"] = self.img_model.has_corrections()
        for (
            correction,
            correction_object,
        ) in self.img_model.img_corrections.corrections.items():
            if correction in ["cbn", "oiadac"]:
                correction_data = correction_object.get_data()
                imcd = corrections_group.create_dataset(
                    correction, correction_data.shape, "f", correction_data
                )
                for param, value in correction_object.get_params().items():
                    imcd.attrs[param] = value
            elif correction == "transfer":
                params = correction_object.get_params()
                transfer_group = corrections_group.create_group("transfer")
                original_data = params["original_data"]
                response_data = params["response_data"]
                original_ds = transfer_group.create_dataset(
                    "original_data", original_data.shape, "f", original_data
                )
                original_ds.attrs["filename"] = params["original_filename"]
                response_ds = transfer_group.create_dataset(
                    "response_data", response_data.shape, "f", response_data
                )
                response_ds.attrs["filename"] = params["response_filename"]

        # the actual image
        image_group.attrs["filename"] = self.img_model.filename
        current_raw_image = self.img_model.untransformed_raw_img_data

        raw_image_data = image_group.create_dataset(
            "raw_image_data", current_raw_image.shape, dtype="f"
        )
        raw_image_data[...] = current_raw_image

        # image transformations
        transformations_group = image_group.create_group("image_transformations")
        for ind, transformation in enumerate(
            self.img_model.get_transformations_string_list()
        ):
            transformations_group.attrs[str(ind)] = transformation

        # save roi data
        if self.roi is not None:
            image_group.attrs["has_roi"] = True
            image_group.create_dataset("roi", (4,), "i8", tuple(self.roi))
        else:
            image_group.attrs["has_roi"] = False

        # save mask model
        mask_group = f.create_group("mask")
        current_mask = self.mask_model.get_mask()
        mask_data = mask_group.create_dataset("data", current_mask.shape, dtype=bool)
        mask_data[...] = current_mask

        # save detector information
        detector_group = f.create_group("detector")
        detector_mode = self.calibration_model.detector_mode
        detector_group.attrs["detector_mode"] = detector_mode.value
        if detector_mode == DetectorModes.PREDEFINED:
            detector_group.attrs["detector_name"] = self.calibration_model.detector.name
        elif detector_mode == DetectorModes.NEXUS:
            detector_group.attrs["nexus_filename"] = (
                self.calibration_model.detector.filename
            )

        # save calibration model
        calibration_group = f.create_group("calibration_model")
        calibration_filename = self.calibration_model.filename
        if calibration_filename.endswith(".poni"):
            base_filename, ext = self.calibration_model.filename.rsplit(".", 1)
        else:
            base_filename = self.calibration_model.filename
            ext = "poni"
        calibration_group.attrs["calibration_filename"] = base_filename + "." + ext
        pyfai_param, _ = self.calibration_model.get_calibration_parameter()
        pfp = calibration_group.create_group("pyfai_parameters")
        for key in pyfai_param:
            try:
                pfp.attrs[key] = pyfai_param[key]
            except TypeError:
                pfp.attrs[key] = ""
        calibration_group.attrs["correct_solid_angle"] = self.correct_solid_angle
        if self.calibration_model.distortion_spline_filename is not None:
            calibration_group.attrs["distortion_spline_filename"] = (
                self.calibration_model.distortion_spline_filename
            )

        # save background pattern and pattern model
        background_pattern_group = f.create_group("background_pattern")
        try:
            background_pattern_x = self.pattern_model.background_pattern._original_x
            background_pattern_y = self.pattern_model.background_pattern._original_y
        except (TypeError, AttributeError):
            background_pattern_x = None
            background_pattern_y = None
        if background_pattern_x is not None and background_pattern_y is not None:
            background_pattern_group.attrs["has_background_pattern"] = True
            bgx = background_pattern_group.create_dataset(
                "x", background_pattern_x.shape, dtype="f"
            )
            bgy = background_pattern_group.create_dataset(
                "y", background_pattern_y.shape, dtype="f"
            )
            bgx[...] = background_pattern_x
            bgy[...] = background_pattern_y
        else:
            background_pattern_group.attrs["has_background_pattern"] = False

        pattern_group = f.create_group("pattern")
        try:
            pattern_x = self.pattern_model.pattern._original_x
            pattern_y = self.pattern_model.pattern._original_y
        except (TypeError, AttributeError):
            pattern_x = None
            pattern_y = None
        if pattern_x is not None and pattern_y is not None:
            px = pattern_group.create_dataset("x", pattern_x.shape, dtype="f")
            py = pattern_group.create_dataset("y", pattern_y.shape, dtype="f")
            px[...] = pattern_x
            py[...] = pattern_y
        pattern_group.attrs["pattern_filename"] = self.pattern_model.pattern_filename
        pattern_group.attrs["unit"] = self.pattern_model.unit
        pattern_group.attrs["file_iteration_mode"] = (
            self.pattern_model.file_iteration_mode
        )
        if self.pattern_model.pattern.auto_bkg:
            pattern_group.attrs["auto_background_subtraction"] = True
            auto_background_group = pattern_group.create_group(
                "auto_background_settings"
            )
            auto_bkg = self.pattern_model.pattern.auto_bkg

            if type(auto_bkg) == SmoothBrucknerBackground:
                auto_background_group.attrs["smoothing"] = auto_bkg.smooth_width
                auto_background_group.attrs["iterations"] = auto_bkg.iterations
                auto_background_group.attrs["poly_order"] = auto_bkg.cheb_order

            auto_bkg_roi = self.pattern_model.pattern.auto_bkg_roi
            if auto_bkg_roi is not None:
                auto_background_group.attrs["x_start"] = auto_bkg_roi[0]
                auto_background_group.attrs["x_end"] = auto_bkg_roi[1]
        else:
            pattern_group.attrs["auto_background_subtraction"] = False

    def load_from_hdf5(self, hdf5_group):
        """
        Loads a configuration from the specified hdf5_group.
        :type hdf5_group: h5py.Group
        """

        f = hdf5_group

        # disable all automatic functions
        self.auto_integrate_pattern = False
        self.auto_integrate_cake = False
        self.auto_save_integrated_pattern = False

        # get working directories
        working_directories = {}
        for key, value in f.get("working_directories").attrs.items():
            if os.path.isdir(value):
                working_directories[key] = value
            else:
                working_directories[key] = ""
        self.working_directories = working_directories

        # load pyFAI parameters
        pyfai_parameters = {}
        for key, value in (
            f.get("calibration_model").get("pyfai_parameters").attrs.items()
        ):
            pyfai_parameters[key] = value

        try:
            self.calibration_model.set_pyFAI(pyfai_parameters)
            filename = f.get("calibration_model").attrs["calibration_filename"]
            (file_path, base_name) = os.path.split(filename)
            self.calibration_model.filename = filename
            self.calibration_model.calibration_name = base_name

        except (KeyError, ValueError):
            print("Problem with saved pyFAI calibration parameters")
            pass

        try:
            self.correct_solid_angle = f.get("calibration_model").attrs[
                "correct_solid_angle"
            ]
        except KeyError:
            pass

        try:
            distortion_spline_filename = f.get("calibration_model").attrs[
                "distortion_spline_filename"
            ]
            self.calibration_model.load_distortion(distortion_spline_filename)
        except KeyError:
            pass

        # load detector definition
        try:
            detector_mode = f.get("detector").attrs["detector_mode"]
            if detector_mode == DetectorModes.PREDEFINED.value:
                detector_name = f.get("detector").attrs["detector_name"]
                self.calibration_model.load_detector(detector_name)
            elif detector_mode == DetectorModes.NEXUS.value:
                nexus_filename = f.get("detector").attrs["nexus_filename"]
                self.calibration_model.load_detector_from_file(nexus_filename)
        except AttributeError:  # to ensure backwards compatibility
            pass

        # load img_model
        self.img_model._img_data = np.copy(
            f.get("image_model").get("raw_image_data")[...]
        )
        filename = f.get("image_model").attrs["filename"]
        self.img_model.filename = filename

        try:
            self.img_model.file_name_iterator.update_filename(filename)
            self.img_model._directory_watcher.path = os.path.dirname(filename)
        except EnvironmentError:
            pass

        self.img_model.autoprocess = f.get("image_model").attrs["auto_process"]
        self.img_model.autoprocess_changed.emit()
        self.img_model.factor = f.get("image_model").attrs["factor"]

        try:
            self.img_model.series_max = f.get("image_model").attrs["series_max"]
            self.img_model.series_pos = f.get("image_model").attrs["series_pos"]
        except KeyError:
            pass

        if f.get("image_model").attrs["has_background"]:
            self.img_model.background_data = np.copy(
                f.get("image_model").get("background_data")[...]
            )
            self.img_model.background_filename = f.get("image_model").attrs[
                "background_filename"
            ]
            self.img_model.background_scaling = f.get("image_model").attrs[
                "background_scaling"
            ]
            self.img_model.background_offset = f.get("image_model").attrs[
                "background_offset"
            ]

        # load image transformations
        transformation_group = f.get("image_model").get("image_transformations")
        transformation_list = []
        for key, transformation in transformation_group.attrs.items():
            transformation_list.append(transformation)
        self.calibration_model.load_transformations_string_list(transformation_list)
        self.img_model.load_transformations_string_list(transformation_list)

        # load roi data
        if f.get("image_model").attrs["has_roi"]:
            self.roi = tuple(f.get("image_model").get("roi")[...])

        # load mask model
        self.mask_model.set_mask(np.copy(f.get("mask").get("data")[...]))

        # load pattern model
        if f.get("pattern").get("x") and f.get("pattern").get("y"):
            self.pattern_model.set_pattern(
                f.get("pattern").get("x")[...],
                f.get("pattern").get("y")[...],
                f.get("pattern").attrs["pattern_filename"],
                f.get("pattern").attrs["unit"],
            )
            self.pattern_model.file_iteration_mode = f.get("pattern").attrs[
                "file_iteration_mode"
            ]
        self.integration_unit = f.get("general_information").attrs["integration_unit"]

        if f.get("background_pattern").attrs["has_background_pattern"]:
            self.pattern_model.background_pattern = Pattern(
                f.get("background_pattern").get("x")[...],
                f.get("background_pattern").get("y")[...],
                "background_pattern",
            )

        if f.get("pattern").attrs["auto_background_subtraction"]:
            self.pattern_model.pattern.auto_bkg_roi = [
                f.get("pattern").get("auto_background_settings").attrs["x_start"],
                f.get("pattern").get("auto_background_settings").attrs["x_end"],
            ]
            self.pattern_model.pattern.auto_bkg = SmoothBrucknerBackground(
                f.get("pattern").get("auto_background_settings").attrs["smoothing"],
                f.get("pattern").get("auto_background_settings").attrs["iterations"],
                f.get("pattern").get("auto_background_settings").attrs["poly_order"],
            )

        # load general configuration
        if f.get("general_information").attrs["integration_num_points"]:
            self.integration_rad_points = int(
                f.get("general_information").attrs["integration_num_points"]
            )

        # cake parameters:
        self.auto_integrate_cake = f.get("general_information").attrs[
            "auto_integrate_cake"
        ]
        try:
            self.cake_azimuth_points = f.get("general_information").attrs[
                "cake_azimuth_points"
            ]
        except KeyError as e:
            pass
        try:
            if f.get("general_information").attrs["cake_azimuth_range"] == "None":
                self.cake_azimuth_range = None
            else:
                self.cake_azimuth_range = f.get("general_information").attrs[
                    "cake_azimuth_range"
                ]
        except KeyError as e:
            pass

        # mask parameters
        self.use_mask = f.get("general_information").attrs["use_mask"]
        self.transparent_mask = f.get("general_information").attrs["transparent_mask"]

        # corrections
        if f.get("image_model").get("corrections").attrs["has_corrections"]:
            for name, correction_group in (
                f.get("image_model").get("corrections").items()
            ):
                params = {}
                for param, val in correction_group.attrs.items():
                    params[param] = val
                if name == "cbn":
                    tth_array = (
                        180.0 / np.pi * self.calibration_model.pattern_geometry.ttha
                    )
                    azi_array = (
                        180.0 / np.pi * self.calibration_model.pattern_geometry.chia
                    )
                    cbn_correction = CbnCorrection(
                        tth_array=tth_array, azi_array=azi_array
                    )

                    cbn_correction.set_params(params)
                    cbn_correction.update()
                    self.img_model.add_img_correction(cbn_correction, name)
                elif name == "oiadac":
                    tth_array = (
                        180.0 / np.pi * self.calibration_model.pattern_geometry.ttha
                    )
                    azi_array = (
                        180.0 / np.pi * self.calibration_model.pattern_geometry.chia
                    )
                    oiadac = ObliqueAngleDetectorAbsorptionCorrection(
                        tth_array=tth_array, azi_array=azi_array
                    )

                    oiadac.set_params(params)
                    oiadac.update()
                    self.img_model.add_img_correction(oiadac, name)
                elif name == "transfer":
                    params = {
                        "original_data": correction_group.get("original_data")[...],
                        "original_filename": correction_group.get(
                            "original_data"
                        ).attrs["filename"],
                        "response_data": correction_group.get("response_data")[...],
                        "response_filename": correction_group.get(
                            "response_data"
                        ).attrs["filename"],
                    }

                    self.img_model.transfer_correction.set_params(params)
                    self.img_model.enable_transfer_function()

        # autosave parameters
        self.auto_save_integrated_pattern = f.get("general_information").attrs[
            "auto_save_integrated_pattern"
        ]
        self.integrated_patterns_file_formats = []
        for file_format in f.get("general_information").get(
            "integrated_patterns_file_formats"
        ):
            self.integrated_patterns_file_formats.append(file_format[0].decode("utf-8"))

        if self.calibration_model.is_calibrated:
            self.integrate_image_1d()
        else:
            self.pattern_model.pattern.recalculate_pattern()



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
from sys import platform as _platform

# If QT_API is not set, use PyQt6 by default
if "QT_API" not in os.environ:
    try:
        import PyQt6.QtCore
    except ImportError:
        pass

from qtpy import QtWidgets
from qt_material import apply_stylesheet

try:
    from pyshortcuts import make_shortcut
except ImportError:
    make_shortcut = None

__version__ = "0.6.0"

from .paths import resources_path, calibrants_path, icons_path, data_path, style_path
from .excepthook import excepthook
from .controller.MainController import MainController


theme_path = os.path.join(style_path, "dark_orange.xml")
qss_path = os.path.join(style_path, "qt_material.css")

def main():
    app = QtWidgets.QApplication([])

    apply_stylesheet(
        app,
        theme=theme_path,
        css_file=qss_path,
        extra={"density_scale": -2},
    )
    sys.excepthook = excepthook
    print("Dioptas {}".format(__version__))

    if len(sys.argv) == 1:  # normal start
        controller = MainController()
        controller.show_window()
        app.exec_()
    else:  # with command line arguments
        if sys.argv[1] == "test":
            controller = MainController(use_settings=False)
            controller.show_window()

        elif sys.argv[1].startswith("makeshortcut"):
            if make_shortcut is None:
                raise ImportError("pyshortcuts not installed.  Try `pip install pyshortcuts`")
            binary_dir = "Scripts" if os.name == "nt" else "bin"
            make_shortcut(
                os.path.join(sys.exec_prefix, binary_dir, "dioptas"),
                name = "Dioptas",
                description="Dioptas 2D XRD {}".format(__version__),
                icon=os.path.join(icons_path, "icon")
                )

        elif sys.argv[1].startswith("version"):
            print(__version__)

        elif sys.argv[1].endswith(".json"):
            controller = MainController(config_file=sys.argv[1])
            controller.show_window()
            app.exec_()
    del app



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .util import Signal
from .util.jcpds import jcpds, jcpds_reflection
from .util.cif import CifConverter
from .util.HelperModule import calculate_color


class PhaseLoadError(Exception):
    def __init__(self, filename):
        super(PhaseLoadError, self).__init__()
        self.filename = filename

    def __repr__(self):
        return "Could not load {0} as jcpds file".format(self.filename)


class PhaseModel(object):

    num_phases = 0

    def __init__(self):
        super(PhaseModel, self).__init__()
        self.phases = []  # type: list[jcpds]
        self.reflections = []
        self.phase_files = []
        self.phase_colors = []
        self.phase_visible = []

        self.same_conditions = True

        self.phase_added = Signal()
        self.phase_removed = Signal(int)  # phase ind
        self.phase_changed = Signal(int)  # phase ind
        self.phase_reloaded = Signal(int)  # phase ind

        self.reflection_added = Signal(int)
        self.reflection_deleted = Signal(int, int)  # phase index, reflection index

    def add_jcpds(self, filename):
        """
        Adds a jcpds file
        :param filename: filename of the jcpds file
        """
        try:
            jcpds_object = jcpds()
            jcpds_object.load_file(filename)
            self.phase_files.append(filename)
            self.add_jcpds_object(jcpds_object)
        except (ZeroDivisionError, UnboundLocalError, ValueError):
            raise PhaseLoadError(filename)

    def add_cif(self, filename, intensity_cutoff=0.5, minimum_d_spacing=0.5):
        """
        Adds a cif file. Internally it is converted to a jcpds format. It calculates the intensities for all of the
        reflections based on the atomic positions
        :param filename: name of the cif file
        :param intensity_cutoff: all reflections added to the jcpds will have larger intensity in % (0-100)
        :param minimum_d_spacing: all reflections added to the jcpds will have larger d spacing than specified here
        """
        try:
            cif_converter = CifConverter(0.31, minimum_d_spacing, intensity_cutoff)
            jcpds_object = cif_converter.convert_cif_to_jcpds(filename)
            self.phase_files.append(filename)
            self.add_jcpds_object(jcpds_object)
        except (ZeroDivisionError, UnboundLocalError, ValueError) as e:
            print(e)
            raise PhaseLoadError(filename)

    def add_jcpds_object(self, jcpds_object):
        """
        Adds a jcpds object to the phase list.
        :param jcpds_object: jcpds object
        :type jcpds_object: jcpds
        """
        self.phases.append(jcpds_object)
        self.reflections.append([])
        self.phase_colors.append(calculate_color(PhaseModel.num_phases + 9))
        self.phase_visible.append(True)
        PhaseModel.num_phases += 1
        if self.same_conditions and len(self.phases) > 2:
            self.phases[-1].compute_d(self.phases[-2].params['pressure'], self.phases[-2].params['temperature'])
        else:
            self.phases[-1].compute_d()
        self.get_lines_d(-1)
        self.phase_added.emit()
        self.phase_changed.emit(len(self.phases) - 1)

    def save_phase_as(self, ind, filename):
        """
        Save the phase specified with ind as a jcpds file.
        """
        self.phases[ind].save_file(filename)
        self.phase_changed.emit(ind)

    def del_phase(self, ind):
        """
        Deletes the a phase with index ind from the phase list
        """
        del self.phases[ind]
        del self.reflections[ind]
        del self.phase_files[ind]
        del self.phase_colors[ind]
        del self.phase_visible[ind]
        self.phase_removed.emit(ind)

    def reload(self, ind):
        """
        Reloads a phase specified by index ind from it's original source filename
        """
        self.clear_reflections(ind)
        self.phases[ind].reload_file()
        for _ in range(len(self.phases[ind].reflections)):
            self.reflection_added.emit(ind)
        self.get_lines_d(ind)
        self.phase_changed.emit(ind)

    def set_pressure(self, ind, pressure):
        """
        Sets the pressure of a phase with index ind. In case same_conditions is true, all phase pressures will be
        updated.
        """
        if self.same_conditions:
            for j in range(len(self.phases)):
                self._set_pressure(j, pressure)
                self.phase_changed.emit(j)
        else:
            self._set_pressure(ind, pressure)
            self.phase_changed.emit(ind)

    def _set_pressure(self, ind, pressure):
        self.phases[ind].compute_d(pressure=pressure)
        self.get_lines_d(ind)

    def set_temperature(self, ind, temperature):
        """
        Sets the temperature of a phase with index ind. In case same_conditions is true, all phase temperatures will be
        updated.
        """
        if self.same_conditions:
            for j in range(len(self.phases)):
                self._set_temperature(j, temperature)
                self.phase_changed.emit(j)
        else:
            self._set_temperature(ind, temperature)
            self.phase_changed.emit(ind)

    def _set_temperature(self, ind, temperature):
        if self.phases[ind].has_thermal_expansion():
            self.phases[ind].compute_d(temperature=temperature)
            self.get_lines_d(ind)

    def set_pressure_temperature(self, ind, pressure, temperature):
        self.phases[ind].compute_d(temperature=temperature, pressure=pressure)
        self.get_lines_d(ind)
        self.phase_changed.emit(ind)

    def set_param(self, ind, param, value):
        """
        Sets one of the jcpds parameters for the phase with index ind to a certain value. Automatically emits the
        phase_changed signal.
        """

        self.phases[ind].params[param] = value
        self.phases[ind].compute_v0()
        self.phases[ind].compute_d0()
        self.phases[ind].compute_d()
        self.get_lines_d(ind)
        self.phase_changed.emit(ind)

    def set_color(self, ind, color):
        """
        Changes the color of the phase with index ind.
        :param ind: index of phase
        :param color: tuple with RGB values (0-255)
        """
        self.phase_colors[ind] = color
        self.phase_changed.emit(ind)

    def set_phase_visible(self, ind, bool):
        """
        Sets the visible flag (bool) for phase with index ind.
        """
        self.phase_visible[ind] = bool
        self.phase_changed.emit(ind)

    def get_lines_d(self, ind):
        """
        Gets the reflections from the phase with index ind and saves them in a two-dimensional array.
        """
        reflections = self.phases[ind].get_reflections()
        res = np.zeros((len(reflections), 5))
        for i, reflection in enumerate(reflections):
            res[i, 0] = reflection.d
            res[i, 1] = reflection.intensity
            res[i, 2] = reflection.h
            res[i, 3] = reflection.k
            res[i, 4] = reflection.l
        self.reflections[ind] = res
        return res

    def get_phase_line_positions(self, ind, unit, wavelength):
        """
        Gets the line positions of phase with index ind in a specfic unit.
        :param ind: phase index
        :param unit: unit for the positions, possible values: '2th_deg', 'q_A^-1', 'd_A'
        :param wavelength: wavelength in nm
        """
        positions = self.reflections[ind][:, 0]
        if unit == 'q_A^-1' or unit == '2th_deg':
            positions = 2 * \
                        np.arcsin(wavelength / (2 * positions)) * 180.0 / np.pi
            if unit == 'q_A^-1':
                positions = 4 * np.pi / wavelength * \
                            np.sin(positions / 360 * np.pi)
        return positions

    def get_phase_line_intensities(self, ind, positions, pattern, x_range, y_range):
        """
        Gets the phase line intensities scaled to each other for a specific x and y range and also a maximum intensity
        based on a specific pattern.
        :param ind: phase index
        :param positions: positions of the lines
        :param pattern: pattern with what it will be plotted
        :param x_range: x range for which the relative intensities will be calculated
        :param y_range: y range for which the relative intensities will be calculated
        :return: array of intensities, baseline representing the start for the lines
        """
        x, y = pattern.data
        if len(y) != 0:
            y_in_range = y[(x > x_range[0]) & (x < x_range[1])]
            if len(y_in_range) == 0:
                return [], 0
            max_pattern_intensity = np.min([np.max(y_in_range), y_range[1]])
        else:
            max_pattern_intensity = 1

        baseline = y_range[0]
        phase_line_intensities = self.reflections[ind][:, 1]
        # search for reflections within current pattern view range
        phase_line_intensities_in_range = phase_line_intensities[(positions > x_range[0]) & (positions < x_range[1])]

        # rescale intensity based on the lines visible
        if len(phase_line_intensities_in_range):
            scale_factor = (max_pattern_intensity - baseline) / \
                           np.max(phase_line_intensities_in_range)
        else:
            scale_factor = 1
        if scale_factor <= 0:
            scale_factor = 0.01

        phase_line_intensities = scale_factor * self.reflections[ind][:, 1] + baseline
        return phase_line_intensities, baseline

    def get_rescaled_reflections(self, ind, pattern, x_range,
                                 y_range, wavelength, unit='2th_deg'):

        """
        Gets the phase line positions and intensities for a phase with index ind scaled to each other for a specific x
        and y range and also a maximum intensity based on a specific pattern.
        :param ind: phase index
        :param pattern: pattern with what it will be plotted
        :param x_range: x range for which the relative intensities will be calculated
        :param y_range: y range for which the relative intensities will be calculated
        :param wavelength: wavelength in nm
        :param unit: unit for the positions, possible values: '2th_deg', 'q_A^-1', 'd_A'
        :return: a tuple with: (array of positions, array of intensities, baseline value)
        """
        positions = self.get_phase_line_positions(ind, unit, wavelength)

        intensities, baseline = self.get_phase_line_intensities(ind, positions, pattern, x_range, y_range)
        return positions, intensities, baseline

    def add_reflection(self, ind):
        """
        Adds an empty reflection to the reflection table of a phase with index ind
        """
        self.phases[ind].add_reflection()
        self.get_lines_d(ind)
        self.reflection_added.emit(ind)

    def delete_reflection(self, phase_ind, reflection_ind):
        """
        Deletes a reflection from a phase with index phase index.
        """
        self.phases[phase_ind].delete_reflection(reflection_ind)
        self.get_lines_d(phase_ind)
        self.reflection_deleted.emit(phase_ind, reflection_ind)
        self.phase_changed.emit(phase_ind)

    def delete_multiple_reflections(self, phase_ind, indices):
        """
        Deletes multiple reflection from a phase with index phase index.
        """
        indices = np.array(sorted(indices))
        for reflection_ind in indices:
            self.delete_reflection(phase_ind, reflection_ind)
            indices -= 1

    def clear_reflections(self, phase_ind):
        """
        Deletes all reflections from a phase with index phase_ind
        """
        for ind in range(len(self.phases[phase_ind].reflections)):
            self.delete_reflection(phase_ind, 0)

    def update_reflection(self, phase_ind, reflection_ind, reflection):
        """
        Updates the reflection of a phase with a new jcpds_reflection
        :param phase_ind: index of the phase
        :param reflection_ind: index of the refection
        :param reflection: updated reflection
        :type reflection: jcpds_reflection
        """
        self.phases[phase_ind].reflections[reflection_ind] = reflection
        self.phases[phase_ind].params['modified'] = True
        self.phases[phase_ind].compute_d0()
        self.phases[phase_ind].compute_d()
        self.get_lines_d(phase_ind)
        self.phase_changed.emit(phase_ind)

    def reset(self):
        """
        Deletes all phases within the phase model.
        """
        for ind in range(len(self.phases)):
            self.del_phase(0)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtWidgets, QtCore

from ...CustomWidgets import (
    IntegerTextField,
    MenuTabWidget,
    NumberTextField,
    LabelAlignRight,
    SpinBoxAlignRight,
    ConservativeSpinBox,
    CheckableFlatButton,
    SaveIconButton,
)


class OptionsWidget(QtWidgets.QWidget):
    def __init__(self):
        super(OptionsWidget, self).__init__()

        self.create_integration_gb()
        self.create_cake_gb()

        self.style_integration_widgets()
        self.style_cake_widgets()
        self.set_tooltips()

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(0, 5, 0, 0)
        self._layout.setSpacing(0)

        self._tab_widget = MenuTabWidget()
        self._tab_widget.add_tab("1D Integration", self.integration_gb)
        self._tab_widget.add_tab("2D Integration", self.cake_gb)

        self._layout.addWidget(self._tab_widget)

        self.setLayout(self._layout)
        self.set_stylesheet()

    def create_integration_gb(self):
        self.integration_gb = QtWidgets.QGroupBox("1D integration")
        self._integration_gb_layout = QtWidgets.QGridLayout()
        self._integration_gb_layout.setContentsMargins(5, 8, 5, 7)
        self._integration_gb_layout.setSpacing(5)
        self.oned_azimuth_min_txt = NumberTextField("-180")
        self.oned_azimuth_max_txt = NumberTextField("180")
        self.oned_full_toggle_btn = CheckableFlatButton("Full")

        self.bin_count_txt = IntegerTextField("0")
        self.bin_count_cb = QtWidgets.QCheckBox("auto")
        self.supersampling_sb = SpinBoxAlignRight()
        self.correct_solid_angle_cb = QtWidgets.QCheckBox("correct Solid Angle")
        self.correct_solid_angle_cb.setChecked(True)

        self._integration_gb_layout.addWidget(LabelAlignRight("Radial bins:"), 0, 0)

        self._integration_gb_layout.addWidget(self.bin_count_txt, 0, 1)
        self._integration_gb_layout.addWidget(self.bin_count_cb, 0, 2)
        self._integration_gb_layout.addWidget(LabelAlignRight("Azimuth range:"), 1, 0)
        self._oned_azi_range_layout = QtWidgets.QHBoxLayout()
        self._oned_azi_range_layout.setContentsMargins(0, 0, 0, 0)
        self._oned_azi_range_layout.setSpacing(5)
        self._oned_azi_range_layout.addWidget(self.oned_azimuth_min_txt)
        self._oned_azi_range_layout.addWidget(LabelAlignRight("-"))
        self._oned_azi_range_layout.addWidget(self.oned_azimuth_max_txt)
        self._integration_gb_layout.addLayout(self._oned_azi_range_layout, 1, 1, 1, 2)
        self._integration_gb_layout.addWidget(self.oned_full_toggle_btn, 1, 3)
        self._integration_gb_layout.addWidget(self.correct_solid_angle_cb, 2, 1)
        self._integration_gb_layout.addWidget(LabelAlignRight("Supersampling:"), 3, 0)
        self._integration_gb_layout.addWidget(self.supersampling_sb, 3, 1)

        self._integration_gb_layout.setRowStretch(0, 0)
        self._integration_gb_layout.setRowStretch(1, 0)
        self._integration_gb_layout.setRowStretch(2, 0)
        self._integration_gb_layout.setRowStretch(5, 1)
        self._integration_gb_layout.setColumnStretch(0, 0)
        self._integration_gb_layout.setColumnStretch(1, 0)
        self._integration_gb_layout.setColumnStretch(2, 0)
        self._integration_gb_layout.setColumnStretch(3, 0)
        self._integration_gb_layout.setColumnStretch(4, 1)

        self.integration_gb.setLayout(self._integration_gb_layout)

    def create_cake_gb(self):
        self.cake_gb = QtWidgets.QGroupBox("2D (Cake-) integration")
        self._cake_gb_layout = QtWidgets.QGridLayout()
        self._cake_gb_layout.setContentsMargins(5, 8, 5, 7)
        self._cake_gb_layout.setSpacing(5)

        self.cake_azimuth_points_sb = ConservativeSpinBox()
        self.cake_azimuth_min_txt = NumberTextField("-180")
        self.cake_azimuth_max_txt = NumberTextField("180")
        self.cake_full_toggle_btn = CheckableFlatButton("Full")
        self.cake_integral_width_sb = ConservativeSpinBox()
        self.cake_save_integral_btn = SaveIconButton()

        self._cake_gb_layout.addWidget(LabelAlignRight("Azimuth bins:"), 0, 0)
        self._cake_gb_layout.addWidget(self.cake_azimuth_points_sb, 0, 1)
        self._cake_gb_layout.addWidget(LabelAlignRight("Azimuth range:"), 1, 0)
        self._azi_range_layout = QtWidgets.QHBoxLayout()
        self._azi_range_layout.setContentsMargins(0, 0, 0, 0)
        self._azi_range_layout.setSpacing(5)
        self._azi_range_layout.addWidget(self.cake_azimuth_min_txt)
        self._azi_range_layout.addWidget(LabelAlignRight("-"))
        self._azi_range_layout.addWidget(self.cake_azimuth_max_txt)
        self._cake_gb_layout.addLayout(self._azi_range_layout, 1, 1, 1, 2)
        self._cake_gb_layout.addWidget(self.cake_full_toggle_btn, 1, 3)
        self._cake_gb_layout.addWidget(LabelAlignRight("Integral Width:"), 2, 0)
        self._cake_gb_layout.addWidget(self.cake_integral_width_sb, 2, 1)
        self._cake_gb_layout.addWidget(self.cake_save_integral_btn, 2, 2)

        self._cake_gb_layout.setRowStretch(0, 0)
        self._cake_gb_layout.setRowStretch(1, 0)
        self._cake_gb_layout.setRowStretch(2, 0)
        self._cake_gb_layout.setRowStretch(3, 1)
        self._cake_gb_layout.setColumnStretch(0, 0)
        self._cake_gb_layout.setColumnStretch(1, 0)
        self._cake_gb_layout.setColumnStretch(2, 0)
        self._cake_gb_layout.setColumnStretch(3, 0)
        self._cake_gb_layout.setColumnStretch(4, 1)

        self.cake_gb.setLayout(self._cake_gb_layout)

    def style_integration_widgets(self):
        self.supersampling_sb.setMinimum(1)
        self.supersampling_sb.setMaximum(20)
        self.supersampling_sb.setSingleStep(1)

        self.bin_count_txt.setEnabled(False)
        self.bin_count_cb.setChecked(True)

        self.oned_full_toggle_btn.setChecked(True)
        self.oned_azimuth_min_txt.setDisabled(True)
        self.oned_azimuth_max_txt.setDisabled(True)

    def style_cake_widgets(self):
        self.cake_azimuth_points_sb.setMinimum(1)
        self.cake_azimuth_points_sb.setMaximum(10000)
        self.cake_azimuth_points_sb.setSingleStep(100)

        self.cake_full_toggle_btn.setChecked(True)
        self.cake_azimuth_min_txt.setDisabled(True)
        self.cake_azimuth_max_txt.setDisabled(True)

        self.cake_integral_width_sb.setMinimum(1)
        self.cake_integral_width_sb.setSingleStep(1)
        self.cake_integral_width_sb.setMaximum(1000000)
        button_width = 25
        button_height = 25
        self.cake_save_integral_btn.setIconSize(QtCore.QSize(15, 15))
        self.cake_save_integral_btn.setWidth(button_width)
        self.cake_save_integral_btn.setHeight(button_height)

    def set_tooltips(self):
        self.cake_full_toggle_btn.setToolTip("Set to full available range")
        self.oned_full_toggle_btn.setToolTip("Set to full available range")
        self.cake_save_integral_btn.setToolTip(
            "Save the tth integral next to the cake image"
        )
        self.cake_integral_width_sb.setToolTip(
            "Sets the width used for the integral plot\nnext to the cake image."
        )

    def set_stylesheet(self):
        self.setStyleSheet(
            """
            QSpinBox, QDoubleSpinBox, QLineEdit {
                width: 60px;
            } 
            """
        )



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .BackgroundController import BackgroundController
from .ImageController import ImageController
from .IntegrationController import IntegrationController
from .phase.JcpdsEditorController import JcpdsEditorController
from .overlay.OverlayController import OverlayController
from .phase.PhaseController import PhaseController
from .phase.PhaseController import PhaseInPatternController
from .phase.PhaseInCakeController import PhaseInCakeController
from .PatternController import PatternController
from .MapController import MapController
from .OptionsController import OptionsController
from .BatchController import BatchController



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib
from collections import OrderedDict
from typing import Optional

from qtpy import QtCore, QtGui, QtWidgets
from pyqtgraph.graphicsItems.GraphicsWidget import GraphicsWidget
from pyqtgraph.graphicsItems.ViewBox import *
from pyqtgraph.graphicsItems.GradientEditorItem import *
from pyqtgraph.graphicsItems.LinearRegionItem import *
from pyqtgraph.graphicsItems.PlotDataItem import *
import pyqtgraph.graphicsItems.GradientEditorItem
from pyqtgraph.Point import Point
import pyqtgraph.functions as fn
import pyqtgraph as pg
import numpy as np
from .ColormapPopup import ColormapPopup
from .NormalizedImageItem import NormalizedImageItem
from ..CustomWidgets import FlatButton
from ... import icons_path, style_path

__all__ = ["HistogramLUTItem"]

# add grey_inverse to the list of color gradients:
pyqtgraph.graphicsItems.GradientEditorItem.Gradients["grey_inverse"] = {
    "ticks": [(0.0, (255, 255, 255, 255)), (1.0, (0, 0, 0, 255))],
    "mode": "rgb",
}

pyqtgraph.graphicsItems.GradientEditorItem.Gradients["jet"] = {
    "ticks": [
        (0.0, (0, 0, 128, 255)),
        (0.1, (0, 0, 255, 255)),
        (0.4, (0, 255, 255, 255)),
        (0.5, (0, 255, 0, 255)),
        (0.6, (255, 255, 0, 255)),
        (0.9, (255, 0, 0, 255)),
        (1.0, (128, 0, 0, 255)),
    ],
    "mode": "rgb",
}

# set the error handling for numpy
np.seterr(divide="ignore", invalid="ignore")


class HistogramLUTItem(GraphicsWidget):
    """
    This is a graphicsWidget which provides controls for adjusting the display of an image.
    Includes:

    - Image histogram
    - Movable region over histogram to select black/white levels
    - Gradient editor to define color lookup table for single-channel images
    """

    sigLookupTableChanged = QtCore.Signal(object)
    sigLevelsChanged = QtCore.Signal(object)
    sigLevelChangeFinished = QtCore.Signal(object)

    def __init__(
            self, image=None, fillHistogram=False, orientation="horizontal", autoLevel=None
    ):
        """
        If *image* (ImageItem) is provided, then the control will be automatically linked to the image and changes to the control will be immediately reflected in the image's appearance.
        By default, the histogram is rendered with a fill. For performance, set *fillHistogram* = False.
        """
        GraphicsWidget.__init__(self)
        self.lut = None
        self.imageItem = None
        self.first_image = True
        self.percentageLevel = False
        self.orientation = orientation
        self.autoLevel = autoLevel
        self._img_data = None

        self.layout = QtWidgets.QGraphicsGridLayout()
        self.setLayout(self.layout)
        self.layout.setContentsMargins(1, 1, 1, 1)
        self.layout.setSpacing(0)

        self.vb = ViewBox()

        self.gradient = GradientEditorItem()
        self.gradient.loadPreset("grey")

        self._normalizationLabel = pg.LabelItem("")

        configurationButton = FlatButton()
        configurationButton.setWidth(30)
        configurationButton.setHeight(30)
        configurationButton.setIcon(
            QtGui.QIcon(str(pathlib.Path(icons_path) / "settings.png"))
        )
        configurationButton.setToolTip("Configure colormap")
        configurationButton.clicked.connect(self._configurationButtonClicked)

        proxy = QtWidgets.QGraphicsProxyWidget()
        proxy.setWidget(configurationButton)

        if orientation == "horizontal":
            self.vb.setMouseEnabled(x=True, y=False)
            self.vb.setMaximumHeight(30)
            self.vb.setMinimumHeight(45)
            self.gradient.setOrientation("top")
            self.region = LogarithmRegionItem([0, 1], LinearRegionItem.Vertical)
            self.layout.addItem(self.vb, 1, 0)
            self.layout.addItem(self.gradient, 0, 0)
            self.layout.addItem(self._normalizationLabel, 1, 1)
            self.layout.addItem(proxy, 0, 1)
            self.gradient.setFlag(self.gradient.ItemStacksBehindParent)
            self.vb.setFlag(self.gradient.ItemStacksBehindParent)
        elif orientation == "vertical":
            self.vb.setMouseEnabled(x=False, y=True)
            self.vb.setMaximumWidth(30)
            self.vb.setMinimumWidth(45)
            self.gradient.setOrientation("right")
            self.region = LogarithmRegionItem([0, 1], LinearRegionItem.Horizontal)
            self.layout.addItem(self.vb, 0, 0)
            self.layout.addItem(self.gradient, 0, 1)
            self.layout.addItem(self._normalizationLabel, 1, 0)
            self.layout.addItem(proxy, 1, 1)

        self.gradient.setFlag(self.gradient.ItemStacksBehindParent)
        self.vb.setFlag(self.gradient.ItemStacksBehindParent)
        self.region.setZValue(1000)
        self.vb.addItem(self.region)
        self.vb.setMenuEnabled(False)

        # self.grid = GridItem()
        # self.vb.addItem(self.grid)

        self.gradient.sigGradientChanged.connect(self.gradientChanged)
        self.region.sigRegionChanged.connect(self.regionChanging)
        self.region.sigRegionChangeFinished.connect(self.regionChanged)

        self.vb.sigRangeChanged.connect(self.viewRangeChanged)
        self.plot = PlotDataItem()
        self.vb.autoRange()
        self.fillHistogram(fillHistogram)
        self.plot.setPen(pg.mkPen(color=(50, 150, 50), size=3))

        self.vb.addItem(self.plot)
        self.autoHistogramRange()

        if image is not None:
            self.setImageItem(image)
            # self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)

        self.vb.mouseClickEvent = self.empty_function
        self.vb.mouseDragEvent = self.empty_function
        self.vb.mouseDoubleClickEvent = self.empty_function
        self.vb.wheelEvent = self.empty_function

    def fillHistogram(self, fill=True, level=0.0, color=(100, 100, 200)):
        if fill:
            self.plot.setFillLevel(level)
            self.plot.setFillBrush(color)
        else:
            self.plot.setFillLevel(None)

            # def sizeHint(self, *args):
            # return QtCore.QSizeF(115, 200)

    def paint(self, p, *args):
        pen = self.region.lines[0].pen
        rgn = self.getLevels()
        if self.orientation == "horizontal":
            p1 = self.vb.mapFromViewToItem(
                self, Point(rgn[0], self.vb.viewRect().center().y())
            )
            p2 = self.vb.mapFromViewToItem(
                self, Point(rgn[1], self.vb.viewRect().center().y())
            )
            gradRect = self.gradient.mapRectToParent(self.gradient.gradRect.rect())
            for pen in [fn.mkPen("k", width=3), pen]:
                p.setPen(pen)
                p.drawLine(p1, gradRect.bottomLeft())
                p.drawLine(p2, gradRect.bottomRight())
                p.drawLine(gradRect.bottomLeft(), gradRect.topLeft())
                p.drawLine(gradRect.bottomRight(), gradRect.topRight())

        elif self.orientation == "vertical":
            p1 = self.vb.mapFromViewToItem(
                self, Point(self.vb.viewRect().center().x(), rgn[0])
            )
            p2 = self.vb.mapFromViewToItem(
                self, Point(self.vb.viewRect().center().x(), rgn[1])
            )
            gradRect = self.gradient.mapRectToParent(self.gradient.gradRect.rect())
            for pen in [fn.mkPen("k", width=3), pen]:
                p.setPen(pen)
                p.drawLine(p1, gradRect.bottomLeft())
                p.drawLine(p2, gradRect.topLeft())
                p.drawLine(gradRect.topLeft(), gradRect.topRight())
                p.drawLine(gradRect.bottomLeft(), gradRect.bottomRight())

    def setHistogramRange(self, mn, mx, padding=0.1):
        """Set the Y range on the histogram plot. This disables auto-scaling."""
        self.vb.enableAutoRange(self.vb.YAxis, False)
        if self.orientation == "horizontal":
            self.vb.setXRange(mn, mx, padding)
        elif self.orientation == "vertical":
            self.vb.setYrange(mn, mx, padding)
            # mn -= d*padding
            # mx += d*padding
            # self.range = [mn,mx]
            # self.updateRange()
            # self.vb.setMouseEnabled(False, True)
            # self.region.setBounds([mn,mx])

    def autoHistogramRange(self):
        """Enable auto-scaling on the histogram plot."""
        self.vb.enableAutoRange(self.vb.XAxis, True)
        self.vb.enableAutoRange(self.vb.YAxis, True)
        # self.range = None
        # self.updateRange()
        # self.vb.setMouseEnabled(False, False)

        # def updateRange(self):
        # self.vb.autoRange()
        # if self.range is not None:
        # self.vb.setYRange(*self.range)
        # vr = self.vb.viewRect()

        # self.region.setBounds([vr.top(), vr.bottom()])

    def setImageItem(self, img):
        self.imageItem = img
        self._updateNormalizationLabel(self.getNormalization())
        img.sigImageChanged.connect(self.imageChanged)
        img.setLookupTable(self.getLookupTable)
        self.regionChanged()
        self.imageChanged()

    def activate(self):
        # we should make sure we are not double connected
        try:
            self.imageItem.sigImageChanged.disconnect(self.imageChanged)
        except TypeError:
            pass
        self.imageItem.sigImageChanged.connect(self.imageChanged)
        self.imageItem.setLookupTable(self.getLookupTable)

    def deactivate(self):
        try:
            self.imageItem.sigImageChanged.disconnect(self.imageChanged)
        except TypeError:
            pass
        self.imageItem.setLookupTable(None)

    def gradientChanged(self):
        if self.imageItem is not None:
            if self.gradient.isLookupTrivial():
                self.imageItem.setLookupTable(None)  # lambda x: x.astype(np.uint8))
            else:
                self.imageItem.setLookupTable(
                    self.getLookupTable
                )  ## send function pointer, not the result

        self.lut = None
        # if self.imageItem is not None:
        # self.imageItem.setLookupTable(self.gradient.getLookupTable(512))
        self.sigLookupTableChanged.emit(self)

    def getLookupTable(self, img=None, n=None, alpha=None):
        if n is None:
            if img.dtype == np.uint8:
                n = 256
            else:
                n = 512
        if self.lut is None:
            self.lut = self.gradient.getLookupTable(n, alpha=alpha)
        return self.lut

    def regionChanged(self):
        self.sigLevelChangeFinished.emit(self)

    def regionChanging(self):
        if self.imageItem is not None:
            self.imageItem.setLevels(np.exp(self.region.getRegion()))
        self.sigLevelsChanged.emit(self)
        self.update()

    def imageChanged(self, autoRange=False, img_data=None):
        self._img_data = None

        if img_data is None:
            img_data = self.imageItem.getData(copy=False)
            if img_data is None:
                return

        left_edges_nonzero, log_hist_nonzero = get_histogram_data(img_data)
        if log_hist_nonzero is None:
            return

        if self.orientation == "horizontal":
            self.plot.setData(left_edges_nonzero, log_hist_nonzero)
        elif self.orientation == "vertical":
            self.plot.setData(log_hist_nonzero, left_edges_nonzero)

    def getImageData(self, copy: bool = True) -> Optional[np.ndarray]:
        """Returns currently displayed image data

        :param copy: False to return internal array, do not modify!
        """
        if isinstance(self.imageItem, NormalizedImageItem):
            return self.imageItem.getData(copy=copy)
        if self._img_data is not None:
            return np.array(self._img_data, copy=copy)
        return None

    def getLevels(self):
        return self.region.getRegion()

    def getExpLevels(self):
        rgn = self.getLevels()
        return np.exp(rgn[0]), np.exp(rgn[1])

    def setLevels(self, mn, mx):
        self.region.setRegion([mn, mx])
        if self.imageItem is not None:
            self.imageItem.setLevels(np.exp(self.region.getRegion()))

    def getNormalization(self) -> str:
        """Returns the current image normalization"""
        if isinstance(self.imageItem, NormalizedImageItem):
            return self.imageItem.getNormalization()
        return "linear"

    def setNormalization(self, normalization: str):
        """Set image current normalization

        This has effect only if the associated image item is a NormalizedImageItem.
        """
        if isinstance(self.imageItem, NormalizedImageItem):
            self.imageItem.setNormalization(normalization)
        self._updateNormalizationLabel(normalization)

    def empty_function(self, *args):
        pass

    def _configurationButtonClicked(self):
        widget = ColormapPopup(parent=self.scene().views()[0])

        widget.setCurrentGradient(self.gradient.saveState())
        widget.setCurrentNormalization(self.getNormalization())
        widget.setRange(*self.getExpLevels())
        widget.setData(data=self.getImageData(copy=False), copy=False)
        widget.sigCurrentGradientChanged.connect(self._configurationGradientChanged)
        widget.sigCurrentNormalizationChanged.connect(self._normalizationChanged)

        def rangeChanged(minimum, maximum):
            self.setLevels(minimum, maximum)
            # Update displayed range since it can differ from the provided one
            widget.setRange(*self.getExpLevels())

        widget.sigRangeChanged.connect(rangeChanged)
        button = self.sender()
        if self.orientation == "horizontal":
            position = button.mapToGlobal(QtCore.QPoint(button.width() + 5, 0))
        else:  # vertical
            widget.adjustSize()  # For retrieving dialog size
            position = button.mapToGlobal(
                QtCore.QPoint(button.width() + 5, button.height())
            ) - QtCore.QPoint(0, widget.frameGeometry().height())
        widget.move(position)
        widget.show()

    def _configurationGradientChanged(self, gradient: dict):
        self.gradient.restoreState(gradient)
        self.gradientChanged()

    def _updateNormalizationLabel(self, normalization: str):
        shortname = NormalizedImageItem.getNormalizationShortname(normalization)
        description = NormalizedImageItem.getNormalizationDescription(
            normalization
        ).capitalize()
        self._normalizationLabel.setText(f"<small>{shortname}</small>")
        self._normalizationLabel.setToolTip(f"{description} colormap normalization")

    def _normalizationChanged(self, normalization: str):
        self._updateNormalizationLabel(normalization)
        if isinstance(self.imageItem, NormalizedImageItem):
            self.imageItem.setNormalization(normalization)
            sender = self.sender()
            if isinstance(sender, ColormapPopup):
                sender.setData(data=self.getImageData(copy=False), copy=False)


class LogarithmRegionItem(LinearRegionItem):
    def __contains__(
            self, values=[0, 1], orientation=None, brush=None, movable=True, bounds=None
    ):
        super(LogarithmRegionItem, self).__init__(
            values, orientation, brush, movable, bounds
        )

    def getRegion(self):
        """Return the values at the edges of the region."""
        # if self.orientation[0] == 'h':
        # r = (self.bounds.top(), self.bounds.bottom())
        # else:
        # r = (self.bounds.left(), self.bounds.right())
        r = [(self.lines[0].value()), (self.lines[1].value())]
        return (min(r), max(r))

    def setRegion(self, rgn):
        """Set the values for the edges of the region.

        ==============   ==============================================
        **Arguments:**
        rgn              A list or tuple of the lower and upper values.
        ==============   ==============================================
        """

        if rgn[0] <= 0:
            rgn[0] = 1
        if rgn[1] <= 0:
            rgn[1] = 1
        rgn = np.log(np.array(rgn))
        if self.lines[0].value() == rgn[0] and self.lines[1].value() == rgn[1]:
            return
        self.blockLineSignal = True
        self.lines[0].setValue(rgn[0])
        self.blockLineSignal = False
        self.lines[1].setValue(rgn[1])
        # self.blockLineSignal = False
        try:  # needed due to changes in the pyqtgraph API
            self.lineMoved(0)
        except TypeError:
            self.lineMoved()
        self.lineMoveFinished()


histogram_cache = OrderedDict()


def get_histogram_data(img_data: np.ndarray) -> tuple:
    """calculates the histogram data necessary for the histogram LUt item"""
    if img_data is None:
        return None, None

    ar_hash = hash(img_data[::13, ::17].tobytes())

    if ar_hash in histogram_cache:
        return histogram_cache[ar_hash]
    else:
        log_data = np.log(img_data)
        log_data = log_data[np.isfinite(log_data)]
        if log_data.size == 0:
            return None, None

        hist, bin_edges = np.histogram(log_data, bins=1500)

        mask_nonzero = hist > 0
        left_edges_nonzero = bin_edges[:-1][mask_nonzero]
        hist_nonzero = hist[mask_nonzero]

        log_hist_nonzero = np.log(hist_nonzero)

        histogram_cache[ar_hash] = (left_edges_nonzero, log_hist_nonzero)
        histogram_cache.move_to_end(ar_hash)
        if len(histogram_cache) > 100:  # keep only 100 histograms in cache
            histogram_cache.popitem(last=False)
        return left_edges_nonzero, log_hist_nonzero



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import time
import threading

import queue

from qtpy import QtCore

from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

from . import Signal


class NewFileInDirectoryWatcher(QtCore.QObject):
    """
    This class watches a given filepath for any new files with a given file extension added to it.

    Typical usage::
        def callback_fcn(path):
            print(path)

        watcher = NewFileInDirectoryWatcher(example_path, file_types = ['.tif', '.tiff'])
        watcher.file_added.connect(callback_fcn)

    """
    _file_added_qt = QtCore.Signal(str)  # used internally for inside of an qt application to avoid thread problems

    def __init__(self, path=None, file_types=None, activate=False):
        """
        :param path: path to folder which will be watched
        :param file_types: list of file types which will be watched for, e.g. ['.tif', '.jpeg']
        :param activate: whether the Watcher will already emit signals
        """
        super(NewFileInDirectoryWatcher, self).__init__()

        if path is None:
            path = os.getcwd()
        self._path = path

        if file_types is None:
            self.file_types = set([])
            self.patterns = '*'
        else:
            self.file_types = set(file_types)
            self.patterns = ['*.' + file_type for file_type in file_types]

        self.event_handler = PatternMatchingEventHandler(self.patterns)
        self.event_handler.on_created = self.on_file_created

        self.active = False
        if activate:
            self.activate()

        self.file_added = Signal(str)  # to be used signal from outside
        self._file_added_qt.connect(self.file_added.emit)
        self.filepath_queue = queue.Queue()

    def on_file_created(self, event):
        """
        Called when a new file is created in the watched directory. This function will be called by the watchdog
        event handle. We check whether the file is fully written by observing whether the file size changes. If the
        file size is not changing within 10ms, we assume that the file is fully written and emit the file_added signal.
        """
        file_path = os.path.abspath(event.src_path)
        file_size = -1
        while file_size != os.stat(file_path).st_size:
            file_size = os.stat(file_path).st_size
            time.sleep(0.01)

        self.filepath_queue.put(os.path.abspath(file_path))

    def activate(self):
        if not self.active:
            self.active = True
            self.queue_thread = threading.Thread(target=self.process_events, daemon=True)
            self.queue_thread.start()
            self._start_observing()

    def deactivate(self):
        if self.active:
            self.active = False
            self._stop_observing()
            self.queue_thread.join()

    def _start_observing(self):
        self.observer = Observer()
        self.observer.schedule(self.event_handler, self.path)
        self.observer.start()

    def _stop_observing(self):
        if self.observer.is_alive():
            self.observer.stop()
            self.observer.join()

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, new_path):
        if self.active:
            self._stop_observing()
        self._path = new_path
        if self.active:
            self._start_observing()

    def process_events(self):
        """continuously check for new files and emit the file_added signal"""
        while self.active:
            try:
                file_path = self.filepath_queue.get(False)  # doesn't block
            except queue.Empty:  # raised when the queue is empty
                time.sleep(0.05)
                continue

            if QtCore.QCoreApplication.instance() is not None:
                self._file_added_qt.emit(file_path)
            else:
                self.file_added.emit(file_path)

    def __del__(self):
        """Stop the observer thread when the object is deleted."""
        self.deactivate()
        self.file_added.clear()



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray data
# Copyright (C) 2019 Clemens Prescher (clemens.prescher@gmail.com)
# Institute for Geology and Mineralogy, University of Cologne
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtCore
import numpy as np
import re

from xypattern import Pattern

from .BatchModel import BatchModel


class MapModel(BatchModel):
    """
    Model for 2D maps from multiple pattern.
    """
    map_changed = QtCore.Signal()
    map_cleared = QtCore.Signal()
    map_problem = QtCore.Signal()
    roi_problem = QtCore.Signal()

    def __init__(self, configuration):
        super(MapModel, self).__init__(configuration)

        self.map = Map()
        self.theta_center = 5.9
        self.theta_range = 0.

        self.rois = []
        self.roi_math = ''

        self.possible_dimensions = []
        self.dimension_index = 0

        # Background for image
        self.bg_image = np.zeros([1920, 1200])

    def reset(self):
        self.reset_data()
        self.map.reset()
        self.reset_rois()
        self.possible_dimensions = []
        self.map_cleared.emit()

    def load_img_map(self, filenames, callback_fn=None):
        self.set_image_files(filenames)
        self.integrate_raw_data(0, len(filenames), 1, use_all=True, callback_fn=callback_fn)

        self.configuration.img_model.blockSignals(True)

        for n in range(self.data.shape[0]):
            self.add_map_point(None, Pattern(self.binning, self.data[n, :]), img_filename=filenames[n])

        self.possible_dimensions = find_possible_dimensions(self.data.shape[0])
        self.map.set_manual_positions(0, 0, 1, 1, self.possible_dimensions[0][0], self.possible_dimensions[0][1], True)

        self.configuration.img_model.blockSignals(False)

    def add_map_point(self, pattern_filename, pattern, position=None, img_filename=None):
        """
        Adds a Point to the map.
        :param pattern_filename: filename of the corresponding map
        :param pattern: Pattern object containing the x and y values of the integrated pattern
        :type pattern: Pattern
        :param position: tuple with x and y position
        :param img_filename: corresponding img filename
        """
        self.map.add_point(pattern_filename, pattern, position, img_filename)

    def add_roi(self, start, end, name=''):
        self.rois.append(Roi(start, end, name))

    def reset_rois(self):
        self.rois = []
        self.roi_math = ''

    def calculate_map_data(self):
        """
        Calculates the ROI math and creates the map image.
        """
        self.map.prepare()

        for point in self.map:
            sum_roi = {}
            for roi in self.rois:
                indices_in_roi = roi.ind_in_roi(point.x_data)
                sum_roi[roi.name] = np.sum(point.y_data[indices_in_roi])

            try:
                current_math = self.calculate_roi_math(sum_roi)
            except SyntaxError:  # needed in case of problem with math
                return

            self.map.set_image_intensity(point.position, current_math)
        self.map_changed.emit()

    def create_simple_summing_roi_math(self):
        """
        Sets the roi_math to be summing of all ROIs.
        """
        self.roi_math = '+'.join([roi.name for roi in self.rois])

    def check_roi_math(self):
        """
        Returns: False if a ROI in the math string is missing from the Rois
        """
        names_in_roi_math = re.findall('([a-zA-Z]+)', self.roi_math)
        for name in names_in_roi_math:
            if name not in [roi.name for roi in self.rois]:
                return False
        return True

    def calculate_roi_math(self, sum_int):
        """
        Evaluates current_roi_math by replacing each ROI name with the sum of the values in that range
        :param sum_int: dictionary with ROI names as key and there respective integral sums as values
        :return: the result of the roi_math equation
        """
        if self.roi_math == '':
            self.create_simple_summing_roi_math()
        current_roi_math = self.roi_math
        for roi_letter in sum_int:
            current_roi_math = current_roi_math.replace(roi_letter, str(sum_int[roi_letter]))
        return eval(current_roi_math)

    def is_empty(self):
        return len(self.map) == 0


class Map:
    def __init__(self):
        self.points = []  # list of MapPoints
        self.sorted_points = []  # list of (points index, x, y)
        self.sorted_map = []  # list of point indices, xs, ys (has a length of 3)

        self.px_per_point_x = 100
        self.px_per_point_y = 100

    def add_point(self, pattern_filename, pattern, position=None, img_filename=None):
        """
        Adds a Point to the map
        :param pattern_filename: filename of the corresponding map
        :param pattern: Pattern object containing the x and y values of the integrated pattern
        :type pattern: Pattern
        :param position: tuple with x and y position
        :param img_filename: corresponding img filename
        """
        self.points.append(MapPoint(pattern_filename, pattern, position, img_filename))

    def prepare(self):
        """
        Prepares the map for inserting intensities
        """
        self._sort_points()
        self._get_map_dimensions()
        self._create_empty_map()

    def _sort_points(self):
        """
        Sorts the current points according to x and y positions and saves them in the sorted_points variable.
        """
        datalist = []
        for ind, point in enumerate(self.points):
            datalist.append([ind, point.position[0], point.position[1]])

        self.sorted_points = sorted(datalist, key=lambda x: (x[1], x[2]))
        self.sorted_map = [[row[i] for row in self.sorted_points] for i in range(len(self.sorted_points[1]))]

    def _get_map_dimensions(self):
        """
        Uses the sorted points and map to estimate minimum x and y position, the differences between points
        """
        self.min_x = self.sorted_map[1][0]
        self.min_y = self.sorted_map[2][0]

        self.num_x = self.sorted_map[2].count(self.min_y)
        self.num_y = self.sorted_map[1].count(self.min_x)

        self.diff_x = self.sorted_points[self.num_y][1] - self.sorted_points[0][1]
        self.diff_y = self.sorted_points[1][2] - self.sorted_points[0][2]

    def _create_empty_map(self):
        """
        Uses the estimated map dimension to calculate
        """
        self.hor_size = self.px_per_point_x * self.num_x
        self.ver_size = self.px_per_point_y * self.num_y

        self.um_per_px_in_x = self.diff_x / self.px_per_point_x
        self.um_per_px_in_y = self.diff_y / self.px_per_point_y

        self.new_image = np.zeros([self.hor_size, self.ver_size])

    def all_positions_defined(self):
        for point in self.points:
            if point.position is None:
                return False
        return True

    def sort_points_by_name(self):
        """
        Returns:
            sorted_datalist: a list of all the map files, sorted by natural filename
        """
        return sorted(self.points, key=lambda point: [int(t) if t.isdigit() else t.lower() for t in
                                                      re.split('(\\d+)', point.pattern_filename)])

    def set_manual_positions(self, min_x, min_y, diff_x, diff_y, num_x, num_y, is_hor_first):
        """
        Args:
            min_x: Horizontal minimum position
            min_y: Vertical minimum position
            diff_x: Step in horizontal
            diff_y: Step in vertical
            num_x: Number of horizontal positions
            num_y: Number of vertical positions
            is_hor_first: True of horizontal changes first between files, False if vertical

        """
        x_grid = np.linspace(min_x, min_x + diff_x * (num_x - 1), num_x)
        y_grid = np.linspace(min_y, min_y + diff_y * (num_y - 1), num_y)

        ind = 0
        if is_hor_first:
            for y in y_grid:
                for x in x_grid:
                    self.points[ind].position = (x, y)
                    ind += 1
        else:
            for x in x_grid:
                for y in y_grid:
                    self.points[ind].position = (x, y)
                    ind += 1

        self.min_x = min_x
        self.min_y = min_y

        self.num_x = num_x
        self.num_y = num_y

        self.diff_x = diff_x
        self.diff_y = diff_y

        self._create_empty_map()
        self.positions_set_manually = True

    def set_image_intensity(self, position, intensity):
        range_hor = self.pos_to_range(position[0], self.min_x, self.px_per_point_x, self.diff_x)
        range_ver = self.pos_to_range(position[1], self.min_y, self.px_per_point_y, self.diff_y)
        self.new_image[range_hor, range_ver] = intensity

    def pos_to_range(self, pos, min_pos, px_per_point, diff_pos):
        """
        Args:
            pos: x or y position of point
            min_pos: minimum x or y value map position
            px_per_point: pixels to draw for each map point in the corresponding direction
            diff_pos: difference in corresponding direction between subsequent map files
        Returns:
            pos_range: a slice with the start and end pixels for drawing the current map file
        """
        range_start = round((pos - min_pos) / diff_pos * px_per_point)
        range_end = round(range_start + px_per_point)
        pos_range = slice(int(range_start), int(range_end))
        return pos_range

    def position_from_xy(self, x, y):
        """gives the position in units for a point clicked in the x, y space"""
        hor = self.min_x + x // self.px_per_point_x * self.diff_x
        ver = self.min_y + y // self.px_per_point_y * self.diff_y
        return hor, ver

    def filenames_from_position(self, pos):
        """
        Gives the filenames for a certain position in the map
        :param pos: tuple horizontal and vertical position
        :return: (pattern_filename, image_filename)
        """
        for point in self.points:
            if abs(float(point.position[0]) - pos[0]) < 2E-4 and \
                    abs(point.position[1] - pos[1]) < 2E-4:
                return point.pattern_filename, point.img_filename
        return None, None
        # dist_sqr = {}
        # for filename, filedata in self.map_model.map_data.items():
        #     dist_sqr[filename] = abs(float(filedata['pos_hor']) - hor) ** 2 + abs(float(filedata['pos_ver']) - ver) ** 2
        #
        # return min(dist_sqr, key=dist_sqr.get)

    def is_empty(self):
        return len(self.points) == 0

    def __getitem__(self, index):
        return self.points[index]

    def __len__(self):
        return len(self.points)

    def reset(self):
        self.points = []
        self.sorted_points = []
        self.sorted_map = []


class MapPoint:
    def __init__(self, pattern_filename, pattern, position=None, img_filename=None):
        """
        Defines a point in a map.
        :param pattern_filename: corresponding pattern filename
        :param pattern: corresponding pattern
        :type pattern: Pattern
        :param position: tuple with the position of the map (x, y)
        :param img_filename: corresponding image filename
        """
        self.pattern_filename = pattern_filename
        self.x_data = pattern.x
        self.y_data = pattern.y
        self.position = position
        self.img_filename = img_filename


class Roi:
    def __init__(self, start, end, name=None):
        """
        Defines a ROI
        :param start: start_value of the ROI
        :param end: end_value of the ROI
        :param name: name of the ROI
        """
        self.start = start
        self.end = end
        self.name = name

    def is_in_roi(self, x):
        """
        whether value x is in the ROI
        :param x: x-value
        :type x: float
        :return: True or False
        :rtype: bool
        """
        return self.start < x < self.end

    def ind_in_roi(self, x_array):
        """
        Gets the indices of a numpy array which are in the ROI
        :param x_array: a numpy array
        :return: list of indices
        """
        return np.where((x_array > self.start) & (x_array < self.end))[0]

    @property
    def center(self):
        return self.range / 2 + self.start

    @property
    def range(self):
        return self.end - self.start


def find_possible_dimensions(num_points):
    dimension_pairs = []
    for n in range(1, int(np.floor(np.sqrt(num_points + 1))) + 1):
        if num_points % n == 0:
            dim1 = n
            dim2 = num_points // n
            dimension_pairs.append((dim1, dim2))
            if dim1 != dim2:
                dimension_pairs.append((dim2, dim1))
    dimension_pairs.sort(key=lambda x: ((x[0]+x[1])/2 - np.sqrt(num_points)) ** 2)
    return dimension_pairs



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from functools import partial
import os

from qtpy import QtWidgets, QtCore, QtGui

from ...CustomWidgets import (
    LabelAlignRight,
    FlatButton,
    CheckableFlatButton,
    DoubleSpinBoxAlignRight,
    VerticalSpacerItem,
    HorizontalSpacerItem,
    ListTableWidget,
    DoubleMultiplySpinBoxAlignRight,
    HorizontalLine,
)
from ...CustomWidgets import NoRectDelegate
from .... import icons_path


class OverlayWidget(QtWidgets.QWidget):
    color_btn_clicked = QtCore.Signal(int, QtWidgets.QWidget)
    show_cb_state_changed = QtCore.Signal(int, bool)
    name_changed = QtCore.Signal(int, str)
    scale_sb_value_changed = QtCore.Signal(int, float)
    offset_sb_value_changed = QtCore.Signal(int, float)

    def __init__(self):
        super(OverlayWidget, self).__init__()

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(5, 5, 5, 5)
        self._layout.setSpacing(5)

        self.button_widget = QtWidgets.QWidget(self)
        self.button_widget.setObjectName("overlay_control_widget")
        self._button_layout = QtWidgets.QVBoxLayout(self.button_widget)
        self._button_layout.setContentsMargins(0, 0, 0, 0)
        self._button_layout.setSpacing(6)

        self.add_btn = FlatButton()
        self.delete_btn = FlatButton()
        self.clear_btn = FlatButton()
        self.move_up_btn = FlatButton()
        self.move_down_btn = FlatButton()

        self._button_layout.addWidget(self.add_btn)
        self._button_layout.addWidget(self.delete_btn)
        self._button_layout.addWidget(HorizontalLine())
        self._button_layout.addSpacerItem(VerticalSpacerItem())
        self._button_layout.addWidget(self.clear_btn)
        self._button_layout.addSpacerItem(VerticalSpacerItem())
        self._button_layout.addWidget(HorizontalLine())
        self._button_layout.addWidget(self.move_up_btn)
        self._button_layout.addWidget(self.move_down_btn)
        self._layout.addWidget(self.button_widget)

        self.parameter_widget = QtWidgets.QWidget(self)
        self._parameter_layout = QtWidgets.QVBoxLayout()
        self._parameter_layout.setContentsMargins(0, 0, 0, 0)
        self._parameter_layout.setSpacing(5)

        self.scale_step_msb = DoubleMultiplySpinBoxAlignRight()
        self.offset_step_msb = DoubleMultiplySpinBoxAlignRight()

        self._step_gb = QtWidgets.QWidget()
        self._step_layout = QtWidgets.QVBoxLayout()
        self._step_layout.setContentsMargins(0, 0, 0, 0)
        self._step_layout.setSpacing(4)
        self._step_layout.addWidget(QtWidgets.QLabel("Scale Step"))
        self._step_layout.addWidget(self.scale_step_msb)
        self._step_layout.addWidget(QtWidgets.QLabel("Offset Step"))
        self._step_layout.addWidget(self.offset_step_msb)
        self._step_gb.setLayout(self._step_layout)
        self._parameter_layout.addWidget(self._step_gb)
        self._parameter_layout.addWidget(HorizontalLine())

        self._waterfall_gb = QtWidgets.QWidget()
        self.waterfall_separation_msb = DoubleMultiplySpinBoxAlignRight()
        self.waterfall_btn = QtWidgets.QPushButton("Waterfall")
        self.waterfall_reset_btn = QtWidgets.QPushButton("Reset")

        self._waterfall_layout = QtWidgets.QVBoxLayout()
        self._waterfall_layout.setContentsMargins(0, 0, 0, 0)
        self._waterfall_layout.setSpacing(4)
        self._waterfall_layout.addWidget(self.waterfall_btn)
        self._waterfall_layout.addWidget(self.waterfall_separation_msb)
        self._waterfall_layout.addWidget(self.waterfall_reset_btn)
        self._waterfall_gb.setLayout(self._waterfall_layout)
        self._parameter_layout.addWidget(self._waterfall_gb)
        self._parameter_layout.addWidget(HorizontalLine())

        self._parameter_layout.addItem(VerticalSpacerItem())

        self.set_as_bkg_btn = QtWidgets.QPushButton("As Bkg")
        self.set_as_bkg_btn.setCheckable(True)
        self._background_layout = QtWidgets.QHBoxLayout()
        self._background_layout.setContentsMargins(0, 0, 0, 0)
        self._background_layout.addSpacerItem(HorizontalSpacerItem())
        self._background_layout.addWidget(self.set_as_bkg_btn)
        self._parameter_layout.addLayout(self._background_layout)

        self.parameter_widget.setLayout(self._parameter_layout)

        self.overlay_tw = ListTableWidget(columns=5)
        self.overlay_tw.setObjectName("overlay_table_widget")
        self.overlay_tw.setHorizontalHeaderLabels(["", "", "Name", "Scale", "Offset"])
        self.overlay_tw.horizontalHeader().setVisible(True)
        self.overlay_tw.horizontalHeader().setStretchLastSection(False)
        self.overlay_tw.horizontalHeader().setSectionResizeMode(
            0, QtWidgets.QHeaderView.Fixed
        )
        self.overlay_tw.horizontalHeader().setSectionResizeMode(
            1, QtWidgets.QHeaderView.Fixed
        )
        self.overlay_tw.horizontalHeader().setSectionResizeMode(
            2, QtWidgets.QHeaderView.Stretch
        )
        self.overlay_tw.horizontalHeader().setSectionResizeMode(
            3, QtWidgets.QHeaderView.ResizeToContents
        )
        self.overlay_tw.horizontalHeader().setSectionResizeMode(
            4, QtWidgets.QHeaderView.ResizeToContents
        )

        self.overlay_tw.setColumnWidth(0, 20)
        self.overlay_tw.setColumnWidth(1, 25)
        self.overlay_tw.cellChanged.connect(self.label_editingFinished)
        self.overlay_tw.setItemDelegate(NoRectDelegate())

        self._layout.addWidget(self.overlay_tw, 10)
        self._layout.addWidget(self.parameter_widget, 0)

        # label for alternative view:
        self.overlay_header_btn = QtWidgets.QPushButton("Overlay")
        self.overlay_header_btn.setObjectName("overlay_header_btn")
        self.overlay_header_btn.setEnabled(False)
        self.overlay_header_btn.setVisible(False)
        self._main_layout = QtWidgets.QVBoxLayout()
        self._main_layout.setContentsMargins(0, 0, 0, 0)
        self._main_layout.setSpacing(0)
        self._header_layout = QtWidgets.QHBoxLayout()
        self._header_layout.addWidget(self.overlay_header_btn)
        self._header_layout.addStretch()
        self._main_layout.addLayout(self._header_layout)
        self._main_layout.addLayout(self._layout)
        self.setLayout(self._main_layout)
        self.style_widgets()
        self.add_tooltips()

        self.show_cbs = []
        self.color_btns = []
        self.scale_sbs = []
        self.offset_sbs = []

    def style_widgets(self):
        icon_size = QtCore.QSize(17, 17)
        self.clear_btn.setIcon(QtGui.QIcon(os.path.join(icons_path, "reset_dark.ico")))
        self.clear_btn.setIconSize(icon_size)

        self.add_btn.setIcon(QtGui.QIcon(os.path.join(icons_path, "open.ico")))
        self.add_btn.setIconSize(icon_size)

        self.delete_btn.setIcon(QtGui.QIcon(os.path.join(icons_path, "delete.png")))
        self.delete_btn.setIconSize(QtCore.QSize(12, 14))

        self.move_up_btn.setIcon(QtGui.QIcon(os.path.join(icons_path, "arrow_up.ico")))
        self.move_up_btn.setIconSize(icon_size)

        self.move_down_btn.setIcon(
            QtGui.QIcon(os.path.join(icons_path, "arrow_down.ico"))
        )
        self.move_down_btn.setIconSize(icon_size)

        def modify_btn_to_icon_size(btn):
            button_height = 25
            button_width = 25
            btn.setFixedSize(button_width, button_height)

        modify_btn_to_icon_size(self.add_btn)
        modify_btn_to_icon_size(self.delete_btn)
        modify_btn_to_icon_size(self.clear_btn)
        modify_btn_to_icon_size(self.move_up_btn)
        modify_btn_to_icon_size(self.move_down_btn)

        step_txt_width = 70
        self.scale_step_msb.setFixedWidth(step_txt_width)
        self.offset_step_msb.setFixedWidth(step_txt_width)
        self.waterfall_separation_msb.setFixedWidth(step_txt_width)

        self.scale_step_msb.setMaximum(10.0)
        self.scale_step_msb.setMinimum(0.01)
        self.scale_step_msb.setValue(0.01)

        self.offset_step_msb.setMaximum(100000.0)
        self.offset_step_msb.setMinimum(0.01)
        self.offset_step_msb.setValue(100.0)

        self.waterfall_separation_msb.setMaximum(100000.0)
        self.waterfall_separation_msb.setMinimum(0.01)
        self.waterfall_separation_msb.setValue(100.0)

        self.set_as_bkg_btn.setStyleSheet("font-size: 11px")
        self.waterfall_btn.setFixedWidth(step_txt_width)
        self.waterfall_reset_btn.setFixedWidth(step_txt_width)
        self.set_as_bkg_btn.setFixedWidth(step_txt_width)

        self.overlay_header_btn.setStyleSheet("border-radius: 0px")

    def add_tooltips(self):
        self.add_btn.setToolTip("Loads Overlay(s) from file(s)")
        self.delete_btn.setToolTip("Removes currently selected overlay")
        self.clear_btn.setToolTip("Removes all overlays")
        self.set_as_bkg_btn.setToolTip("Set selected overlay as background")
        self.waterfall_reset_btn.setToolTip("Reset waterfall separation")
        self.waterfall_btn.setToolTip("Apply waterfall separation")

    def add_overlay(self, name, color):
        current_rows = self.overlay_tw.rowCount()
        self.overlay_tw.setRowCount(current_rows + 1)
        self.overlay_tw.blockSignals(True)

        show_cb = QtWidgets.QCheckBox()
        show_cb.setChecked(True)
        show_cb.stateChanged.connect(partial(self.show_cb_changed, show_cb))
        show_cb.setStyleSheet("background-color: transparent")
        self.overlay_tw.setCellWidget(current_rows, 0, show_cb)
        self.show_cbs.append(show_cb)

        color_button = FlatButton()
        color_button.setStyleSheet(f"background-color: {color}; margin: 2px")
        color_button.clicked.connect(partial(self.color_btn_click, color_button))
        self.overlay_tw.setCellWidget(current_rows, 1, color_button)
        self.color_btns.append(color_button)

        name_item = QtWidgets.QTableWidgetItem(name)
        name_item.setFlags(name_item.flags() & ~QtCore.Qt.ItemIsEditable)
        self.overlay_tw.setItem(current_rows, 2, QtWidgets.QTableWidgetItem(name))

        scale_sb = DoubleSpinBoxAlignRight()
        scale_sb.setFixedWidth(70)
        scale_sb.setMinimum(-9999999)
        scale_sb.setMaximum(9999999)
        scale_sb.setValue(1)
        scale_sb.setSingleStep(self.scale_step_msb.value())
        scale_sb.valueChanged.connect(partial(self.scale_sb_callback, scale_sb))
        self.overlay_tw.setCellWidget(current_rows, 3, scale_sb)
        self.scale_sbs.append(scale_sb)

        offset_sb = DoubleSpinBoxAlignRight()
        offset_sb.setFixedWidth(80)
        offset_sb.setMinimum(-9999999)
        offset_sb.setMaximum(9999999)
        offset_sb.setValue(0)
        offset_sb.setSingleStep(self.offset_step_msb.value())
        offset_sb.valueChanged.connect(partial(self.offset_sb_callback, offset_sb))
        self.overlay_tw.setCellWidget(current_rows, 4, offset_sb)
        self.offset_sbs.append(offset_sb)

        self.overlay_tw.setRowHeight(current_rows, 25)
        self.update_overlay_tw_column_sizes()
        self.select_overlay(current_rows)
        self.overlay_tw.blockSignals(False)

    def update_overlay_tw_column_sizes(self):
        self.overlay_tw.setColumnWidth(0, 20)
        self.overlay_tw.setColumnWidth(1, 25)

        self.overlay_tw.horizontalHeader().setSectionResizeMode(
            3, QtWidgets.QHeaderView.ResizeToContents
        )
        self.overlay_tw.horizontalHeader().setSectionResizeMode(
            4, QtWidgets.QHeaderView.ResizeToContents
        )

    def select_overlay(self, ind):
        if self.overlay_tw.rowCount() > 0:
            self.overlay_tw.selectRow(ind)

    def get_selected_overlay_row(self):
        selected = self.overlay_tw.selectionModel().selectedRows()
        try:
            row = selected[0].row()
        except IndexError:
            row = -1
        return row

    def remove_overlay(self, ind):
        self.overlay_tw.blockSignals(True)
        self.overlay_tw.removeRow(ind)
        self.overlay_tw.blockSignals(False)
        del self.show_cbs[ind]
        del self.color_btns[ind]
        del self.offset_sbs[ind]
        del self.scale_sbs[ind]

        if self.overlay_tw.rowCount() > ind:
            self.select_overlay(ind)
        else:
            self.select_overlay(self.overlay_tw.rowCount() - 1)

    def color_btn_click(self, button):
        self.color_btn_clicked.emit(self.color_btns.index(button), button)

    def show_cb_changed(self, checkbox):
        self.show_cb_state_changed.emit(
            self.show_cbs.index(checkbox), checkbox.isChecked()
        )

    def show_cb_set_checked(self, ind, state):
        checkbox = self.show_cbs[ind]
        checkbox.setChecked(state)

    def show_cb_is_checked(self, ind):
        checkbox = self.show_cbs[ind]
        return checkbox.isChecked()

    def label_editingFinished(self, row, col):
        label_item = self.overlay_tw.item(row, col)
        if label_item is not None:
            self.name_changed.emit(row, str(label_item.text()))

    def set_overlay_name(self, ind, label):
        label_item = self.overlay_tw.item(ind, 2)
        if label_item is not None:
            label_item.setText(label)

    def scale_sb_callback(self, scale_sb):
        self.scale_sb_value_changed.emit(
            self.scale_sbs.index(scale_sb), scale_sb.value()
        )

    def offset_sb_callback(self, offset_sb):
        self.offset_sb_value_changed.emit(
            self.offset_sbs.index(offset_sb), offset_sb.value()
        )



# img_px_size in mm
gsecars_photo = {'img_px_size_hor': 0.00035,
                 'img_px_size_ver': 0.00035,
                 'img_hor_px': 1920,
                 'img_ver_px': 1200,
                 'flip_hor_prefixes': 'us,ms',
                 'flip_hor': False,
                 'flip_ver': True,
                 }



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtWidgets, QtCore
from pyqtgraph import GraphicsLayoutWidget
from dioptas.widgets.plot_widgets import PatternWidget
from dioptas.widgets.plot_widgets.ImgWidget import IntegrationImgWidget

from .integration.CustomWidgets import MouseUnitCurrentAndClickedWidget


class MapWidget(QtWidgets.QWidget):
    """
    Defines the main structure of the Map widget, which is separated into several Parts
    """

    def __init__(self, *args, **kwargs):
        super(MapWidget, self).__init__(*args, **kwargs)

        self.create_widgets()
        self.create_layout()

    def create_widgets(self):
        self.map_pg_layout = GraphicsLayoutWidget()
        self.map_plot_widget = IntegrationImgWidget(
            self.map_pg_layout, orientation="horizontal"
        )

        self.img_pg_layout = GraphicsLayoutWidget()
        self.img_plot_widget = IntegrationImgWidget(
            self.img_pg_layout, orientation="horizontal"
        )
        self.map_plot_control_widget = (
            MapPlotControlWidget()
        )  # widget below the map image

        self.pattern_pg_layout = GraphicsLayoutWidget()
        self.pattern_plot_widget = PatternWidget(self.pattern_pg_layout)
        self.pattern_plot_widget.show_map_interactive_roi()

        self.pattern_footer_widget = PatternFooterWidget()
        self.pattern_widget = QtWidgets.QWidget()

        self.control_widget = MapControlWidget()

    def create_layout(self):
        self._outer_layout = TightHBoxLayout()
        self._left_layout = TightVBoxLayout()
        self._right_layout = TightVBoxLayout()
        self._upper_right_layout = TightHBoxLayout()

        self._left_widget = QtWidgets.QWidget()
        self._left_widget.setLayout(self._left_layout)
        self._left_layout.addWidget(self.map_pg_layout)
        self._left_layout.addWidget(self.map_plot_control_widget)

        self.upper_right_splitter = QtWidgets.QSplitter()
        self.upper_right_splitter.setOrientation(QtCore.Qt.Horizontal)
        self.upper_right_splitter.addWidget(self.img_pg_layout)
        self.upper_right_splitter.addWidget(self.control_widget)

        self._lower_right_layout = TightVBoxLayout()
        self._lower_right_layout.addWidget(self.pattern_pg_layout)
        self._lower_right_layout.addWidget(self.pattern_footer_widget)
        self.pattern_widget.setLayout(self._lower_right_layout)

        self.vertical_splitter = QtWidgets.QSplitter(self)
        self.vertical_splitter.setOrientation(QtCore.Qt.Vertical)
        self.vertical_splitter.addWidget(self.upper_right_splitter)
        self.vertical_splitter.addWidget(self.pattern_widget)

        self.horizontal_splitter = QtWidgets.QSplitter()
        self.horizontal_splitter.setOrientation(QtCore.Qt.Horizontal)
        self.horizontal_splitter.addWidget(self._left_widget)
        self.horizontal_splitter.addWidget(self.vertical_splitter)

        self._outer_layout.addWidget(self.horizontal_splitter)

        self.setLayout(self._outer_layout)


class TightHBoxLayout(QtWidgets.QHBoxLayout):
    def __init__(self, *args, **kwargs):
        super(TightHBoxLayout, self).__init__(*args, **kwargs)
        self.setContentsMargins(0, 0, 0, 0)
        self.setSpacing(0)


class TightVBoxLayout(QtWidgets.QVBoxLayout):
    def __init__(self, *args, **kwargs):
        super(TightVBoxLayout, self).__init__(*args, **kwargs)
        self.setContentsMargins(0, 0, 0, 0)
        self.setSpacing(0)


class MapControlWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(MapControlWidget, self).__init__(*args, **kwargs)

        self.create_widgets()
        self.create_layout()
        self.style_widgets()

    def create_widgets(self):
        self.load_btn = QtWidgets.QPushButton("Load")
        self.file_list = QtWidgets.QListWidget()

    def create_layout(self):
        self._outer_layout = TightVBoxLayout()
        self._outer_layout.setContentsMargins(0, 0, 0, 0)
        self._outer_layout.setSpacing(5)

        self._outer_layout.addWidget(self.load_btn)
        self._outer_layout.addWidget(self.file_list)

        self.setLayout(self._outer_layout)

    def style_widgets(self):
        pass


class MapPlotControlWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(MapPlotControlWidget, self).__init__(*args, **kwargs)

        self.create_widgets()
        self.create_layout()
        self.style_widgets()

    def create_widgets(self):
        self.map_dimension_cb = QtWidgets.QComboBox()
        self.mouse_x_label = QtWidgets.QLabel("X: ")
        self.mouse_y_label = QtWidgets.QLabel("Y: ")
        self.mouse_int_label = QtWidgets.QLabel("I: ")
        self.filename_label = QtWidgets.QLabel("")

    def create_layout(self):
        self._outer_layout = TightHBoxLayout()

        self._left_layout = TightVBoxLayout()
        self._mouse_pos_layout = TightHBoxLayout()
        self._mouse_pos_layout.addWidget(self.mouse_x_label)
        self._mouse_pos_layout.addWidget(self.mouse_y_label)
        self._mouse_pos_layout.addWidget(self.mouse_int_label)
        self._left_layout.addLayout(self._mouse_pos_layout)
        self._left_layout.addWidget(self.filename_label)
        self._outer_layout.addWidget(QtWidgets.QLabel("Dim: "))
        self._outer_layout.addWidget(self.map_dimension_cb)
        self._outer_layout.addStretch(1)
        self._outer_layout.addLayout(self._left_layout)
        self.setLayout(self._outer_layout)

    def style_widgets(self):
        self._outer_layout.setContentsMargins(6, 3, 0, 0)
        self.mouse_x_label.setFixedWidth(50)
        self.mouse_y_label.setFixedWidth(50)
        self.mouse_int_label.setMinimumWidth(80)
        self.setMinimumHeight(30)


class PatternFooterWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(PatternFooterWidget, self).__init__(*args, **kwargs)

        self.create_widgets()
        self.create_layout()
        self.style_widgets()

    def create_widgets(self):
        self.mouse_unit_widget = MouseUnitCurrentAndClickedWidget()

    def create_layout(self):
        self._outer_layout = TightHBoxLayout()
        self._outer_layout.addWidget(self.mouse_unit_widget)
        self._outer_layout.addStretch(1)
        self.setLayout(self._outer_layout)

    def style_widgets(self):
        self._outer_layout.setContentsMargins(6, 3, 6, 0)
        self.setMinimumHeight(30)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# imports for type hinting in PyCharm -- DO NOT DELETE
from ...widgets.integration import IntegrationWidget
from ...model.DioptasModel import DioptasModel


class OptionsController(object):
    """
    IntegrationPatternController handles all the interaction from the IntegrationView with the pattern data.
    It manages the auto integration of image files to  in addition to pattern browsing and changing of units
    (2 Theta, Q, A)
    """

    def __init__(self, widget, dioptas_model):
        """
        :param widget: Reference to an IntegrationWidget
        :param dioptas_model: reference to DioptasModel object

        :type widget: IntegrationWidget
        :type dioptas_model: DioptasModel
        """

        self.integration_widget = widget
        self.options_widget = self.integration_widget.integration_control_widget.integration_options_widget

        self.model = dioptas_model

        self.connect_signals()

    def connect_signals(self):
        self.options_widget.correct_solid_angle_cb.stateChanged.connect(self.correct_solid_angle_cb_clicked)
        self.model.configuration_selected.connect(self.update_gui)
        self.model.pattern_changed.connect(self.update_gui)

        self.options_widget.cake_azimuth_points_sb.valueChanged.connect(self.cake_azimuth_points_changed)
        self.options_widget.cake_azimuth_min_txt.editingFinished.connect(self.cake_azimuth_range_changed)
        self.options_widget.cake_azimuth_max_txt.editingFinished.connect(self.cake_azimuth_range_changed)

        self.options_widget.oned_full_toggle_btn.toggled.connect(self.oned_full_toggled_btn_changed)
        self.options_widget.cake_full_toggle_btn.toggled.connect(self.cake_full_toggled_btn_changed)
        self.options_widget.oned_azimuth_min_txt.editingFinished.connect(self.oned_azimuth_range_changed)
        self.options_widget.oned_azimuth_max_txt.editingFinished.connect(self.oned_azimuth_range_changed)

    def correct_solid_angle_cb_clicked(self):
        self.model.current_configuration.correct_solid_angle = self.options_widget.correct_solid_angle_cb.isChecked()

    def update_gui(self):
        self.options_widget.blockSignals(True)

        self.options_widget.correct_solid_angle_cb.blockSignals(True)
        self.options_widget.correct_solid_angle_cb.setChecked(int(self.model.current_configuration.correct_solid_angle))
        self.options_widget.correct_solid_angle_cb.blockSignals(False)

        self.options_widget.bin_count_txt.blockSignals(True)
        self.options_widget.bin_count_txt.setText("{:1.0f}".format(self.model.calibration_model.num_points))
        self.options_widget.bin_count_txt.blockSignals(False)

        self.options_widget.cake_azimuth_points_sb.blockSignals(True)
        self.options_widget.cake_azimuth_points_sb.setValue(self.model.current_configuration.cake_azimuth_points)
        self.options_widget.cake_azimuth_points_sb.blockSignals(False)

        if self.model.current_configuration.cake_azimuth_range is None:
            self.enable_full_cake_range()
        else:
            self.options_widget.cake_azimuth_min_txt.setText(
                '{}'.format(self.model.current_configuration.cake_azimuth_range[0]))
            self.options_widget.cake_azimuth_max_txt.setText(
                '{}'.format(self.model.current_configuration.cake_azimuth_range[1]))
            self.options_widget.blockSignals(False)
            self.disable_full_cake_range()
        self.options_widget.blockSignals(False)

    def cake_azimuth_range_changed(self):
        range_min = float(self.options_widget.cake_azimuth_min_txt.text())
        range_max = float(self.options_widget.cake_azimuth_max_txt.text())
        self.model.current_configuration.cake_azimuth_range = (range_min, range_max)

    def cake_azimuth_points_changed(self):
        self.model.current_configuration.cake_azimuth_points = int(
            self.options_widget.cake_azimuth_points_sb.value())

    def cake_full_toggled_btn_changed(self):
        if self.options_widget.cake_full_toggle_btn.isChecked():
            self.enable_full_cake_range()
            self.model.current_configuration.cake_azimuth_range = None

        elif not self.options_widget.cake_full_toggle_btn.isChecked():
            self.disable_full_cake_range()
            self.cake_azimuth_range_changed()

    def enable_full_cake_range(self):
        self.options_widget.cake_azimuth_min_txt.setDisabled(True)
        self.options_widget.cake_azimuth_max_txt.setDisabled(True)
        self.integration_widget.cake_shift_azimuth_sl.setDisabled(False)

    def disable_full_cake_range(self):
        self.options_widget.cake_azimuth_min_txt.setDisabled(False)
        self.options_widget.cake_azimuth_max_txt.setDisabled(False)
        self.integration_widget.cake_shift_azimuth_sl.setDisabled(True)
        self.integration_widget.cake_shift_azimuth_sl.setValue(0)

    def oned_azimuth_range_changed(self):
        range_min = float(self.options_widget.oned_azimuth_min_txt.text())
        range_max = float(self.options_widget.oned_azimuth_max_txt.text())
        self.model.current_configuration.oned_azimuth_range = (range_min, range_max)

    def oned_full_toggled_btn_changed(self):
        if self.options_widget.oned_full_toggle_btn.isChecked():
            self.enable_full_oned_range()
            self.model.current_configuration.oned_azimuth_range = None

        elif not self.options_widget.oned_full_toggle_btn.isChecked():
            self.disable_full_oned_range()
            self.oned_azimuth_range_changed()

    def enable_full_oned_range(self):
        self.options_widget.oned_azimuth_min_txt.setDisabled(True)
        self.options_widget.oned_azimuth_max_txt.setDisabled(True)

    def disable_full_oned_range(self):
        self.options_widget.oned_azimuth_min_txt.setDisabled(False)
        self.options_widget.oned_azimuth_max_txt.setDisabled(False)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtCore, QtWidgets
import numpy as np
try:
    import epics
except ImportError:
    epics = None

from .econfig import epics_config

from ...widgets.integration import IntegrationWidget
from ...model.DioptasModel import DioptasModel


class EpicsController(object):

    def __init__(self, widget, dioptas_model):
        """
        :param widget: Reference to IntegrationWidget
        :param dioptas_model: Reference to DioptasModel object

        :type widget: IntegrationWidget
        :type dioptas_model: DioptasModel
        """

        self.widget = widget
        self.model = dioptas_model

        self.move_widget = widget.move_widget

        # Create timer
        self.epics_update_timer = QtCore.QTimer(self.widget)

        # read epics_config
        self.hor_motor_name = epics_config['sample_position_x']
        self.ver_motor_name = epics_config['sample_position_y']
        self.focus_motor_name = epics_config['sample_position_z']
        self.omega_motor_name = epics_config['sample_position_omega']

        self.connect_signals()

    def connect_signals(self):
        self.epics_update_timer.timeout.connect(self.update_current_motor_position)

        self.model.img_changed.connect(self.update_image_position)

        self.widget.move_widget_btn.clicked.connect(self.widget.move_widget.raise_widget)
        self.widget.move_widget.connect_epics_btn.clicked.connect(self.update_current_motor_position)
        self.widget.move_widget.move_btn.clicked.connect(self.move_stage)
        self.widget.move_widget.motors_setup_btn.clicked.connect(self.open_motors_setup_widget)

        self.widget.move_widget.motors_setup_widget.reread_config_btn.clicked.connect(self.reread_config)
        self.widget.move_widget.motors_setup_widget.set_motor_names_btn.clicked.connect(self.get_motors)

    def update_current_motor_position(self):
        hor = epics.caget(self.hor_motor_name + '.RBV', as_string=True)
        ver = epics.caget(self.ver_motor_name + '.RBV', as_string=True)
        focus = epics.caget(self.focus_motor_name + '.RBV', as_string=True)
        omega = epics.caget(self.omega_motor_name + '.RBV', as_string=True)

        if ver is not None and hor is not None and focus is not None and omega is not None:
            self.epics_update_timer.start(1000)
        else:
            if self.epics_update_timer.isActive():
                self.epics_update_timer.stop()

        self.move_widget.hor_lbl.setText(str(hor))
        self.move_widget.ver_lbl.setText(str(ver))
        self.move_widget.focus_lbl.setText(str(focus))
        self.move_widget.omega_lbl.setText(str(omega))

    def update_image_position(self):
        try:
            self.move_widget.img_hor_lbl.setText('{:.3f}'.format(self.model.img_model.motors_info['Horizontal']))
            self.move_widget.img_ver_lbl.setText('{:.3f}'.format(self.model.img_model.motors_info['Vertical']))
            self.move_widget.img_focus_lbl.setText('{:.3f}'.format(self.model.img_model.motors_info['Focus']))
            self.move_widget.img_omega_lbl.setText('{:.3f}'.format(self.model.img_model.motors_info['Omega']))
        except KeyError:
            self.move_widget.img_hor_lbl.setText("")
            self.move_widget.img_ver_lbl.setText("")
            self.move_widget.img_focus_lbl.setText("")
            self.move_widget.img_omega_lbl.setText("")

    def move_stage(self):
        hor_pos = float(self.move_widget.img_hor_lbl.text())
        ver_pos = float(self.move_widget.img_ver_lbl.text())
        focus_pos = float(self.move_widget.img_focus_lbl.text())
        omega_pos =float(self.move_widget.img_omega_lbl.text())

        if self.check_sample_point_distances(hor_pos, ver_pos, focus_pos):
            if self.move_widget.move_hor_cb.isChecked():
                epics.caput(self.hor_motor_name + '.VAL', hor_pos)
            if self.move_widget.move_ver_cb.isChecked():
                epics.caput(self.ver_motor_name + '.VAL', ver_pos)
            if self.move_widget.move_focus_cb.isChecked():
                epics.caput(self.focus_motor_name + '.VAL', focus_pos)
            if self.move_widget.move_omega_cb.isChecked():
                if self.check_conditions() is False:
                    self.show_error_message_box(
                        'If you want to rotate the stage, please move mirrors and microscope in the right positions!')
                    return
                elif float(omega_pos) > -45.0 or float(omega_pos) < -135.0:
                    self.show_error_message_box('Requested omega angle is not within the limits')
                    return
                else:
                    epics.caput(self.omega_motor_name + '.VAL', omega_pos)

    def open_motors_setup_widget(self):
        self.move_widget.motors_setup_widget.hor_motor_txt.setText(self.hor_motor_name)
        self.move_widget.motors_setup_widget.ver_motor_txt.setText(self.ver_motor_name)
        self.move_widget.motors_setup_widget.focus_motor_txt.setText(self.focus_motor_name)
        self.move_widget.motors_setup_widget.omega_motor_txt.setText(self.omega_motor_name)
        self.move_widget.motors_setup_widget.show()

    def get_motors(self):
        self.hor_motor_name, self.ver_motor_name, self.focus_motor_name, self.omega_motor_name = \
            self.move_widget.motors_setup_widget.return_motor_names()
        self.update_current_motor_position()

    def reread_config(self):
        self.move_widget.motors_setup_widget.hor_motor_txt.setText(epics_config['sample_position_x'])
        self.move_widget.motors_setup_widget.ver_motor_txt.setText(epics_config['sample_position_y'])
        self.move_widget.motors_setup_widget.focus_motor_txt.setText(epics_config['sample_position_z'])
        self.move_widget.motors_setup_widget.omega_motor_txt.setText(epics_config['sample_position_omega'])

    def closeEvent(self, QCloseEvent):
        self.epics_update_timer.stop()


    @staticmethod
    def show_error_message_box(msg):
        msg_box = QtWidgets.QMessageBox()
        msg_box.setWindowFlags(QtCore.Qt.Tool)
        msg_box.setText(msg)
        msg_box.setIcon(QtWidgets.QMessageBox.Critical)
        msg_box.setWindowTitle('Error')
        msg_box.setStandardButtons(QtWidgets.QMessageBox.Ok)
        msg_box.setDefaultButton(QtWidgets.QMessageBox.Ok)
        msg_box.exec_()

    @staticmethod
    def show_continue_abort_message_box(msg):
        msg_box = QtWidgets.QMessageBox()
        msg_box.setWindowFlags(QtCore.Qt.Tool)
        msg_box.setText("<p align='center' style='font-size:20px' >{}</p>".format(msg))
        msg_box.setIcon(QtWidgets.QMessageBox.Critical)
        msg_box.setWindowTitle('Continue?')
        msg_box.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.Abort)
        msg_box.setDefaultButton(QtWidgets.QMessageBox.Abort)
        msg_box.exec_()
        return msg_box.result()

    @staticmethod
    def check_conditions():
        if int(epics.caget('13IDD:m24.RBV')) > -105:
            return False
        elif int(epics.caget('13IDD:m23.RBV')) > -105:
            return False
        elif int(epics.caget('13IDD:m67.RBV')) > -65:
            return False
        return True

    def check_sample_point_distances(self, pos_x, pos_y, pos_z):
        cur_x = float(epics.caget(self.hor_motor_name + '.RBV', as_string=True))
        cur_y = float(epics.caget(self.ver_motor_name + '.RBV', as_string=True))
        cur_z = float(epics.caget(self.focus_motor_name + '.RBV', as_string=True))

        largest_distance = np.sqrt((float(pos_x) - cur_x) ** 2 + (float(pos_y) - cur_y) ** 2 + (float(pos_z) - cur_z) ** 2)

        if largest_distance > 0.2:
            reply = self.show_continue_abort_message_box(
                'Current and image positions are far away from each other. Are you sure, you want to move?')
            if reply == QtWidgets.QMessageBox.Abort:
                return False

        return True




# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import time
import traceback

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

from .widgets.UtilityWidgets import ErrorMessageBox
from . import __version__


def excepthook(exc_type, exc_value, traceback_obj):
    """
    Global function to catch unhandled exceptions. This function will result in an error dialog which displays the
    error information.

    :param exc_type: exception type
    :param exc_value: exception value
    :param traceback_obj: traceback object
    :return:
    """
    separator = '-' * 80
    log_path = f"{os.path.expanduser('~')}/dioptas_error.log"
    notice = \
        """An unhandled exception occurred. Please report the bug under:\n """ \
        """\t%s\n""" \
        """or via email to:\n\t <%s>.\n\n""" \
        """Please make sure to report the steps to reproduce the error. Otherwise it will be hard to fix it. \n\n""" \
        """A log has been written to "%s".\n\nError information:\n""" % \
        ("https://github.com/Dioptas/Dioptas/issues",
         "clemens.prescher@gmail.com", log_path)
    version_info = '\n'.join((separator, "Dioptas Version: %s" % __version__))
    time_string = time.strftime("%Y-%m-%d, %H:%M:%S")
    tb_info_file = StringIO()
    traceback.print_tb(traceback_obj, None, tb_info_file)
    tb_info_file.seek(0)
    tb_info = tb_info_file.read()
    errmsg = '%s: \n%s' % (str(exc_type), str(exc_value))
    sections = [separator, time_string, separator, errmsg, separator, tb_info]
    msg = '\n'.join(sections)
    try:
        f = open(log_path, "a")
        f.write(msg)
        f.write(version_info)
        f.close()
    except IOError:
        pass
    errorbox = ErrorMessageBox()
    errorbox.setText(str(notice) + str(msg) + str(version_info))
    errorbox.exec_()
